"""PIN 10 methods."""

# Import system modules
import os
import sys
sys.path.insert(1, '..' + os.sep + 'src')


from hardware.exceptions import InvalidPin10Value  # noqa


def read_pin10(self) -> int:
    """
    Return the value of PIN 10 on the i4004 chip (simulated test pin).

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    PIN_10_SIGNAL_TEST
        The value of the simulated test pin

    """
    return self.PIN_10_SIGNAL_TEST


def write_pin10(self, value: int) -> bool:
    """
    Write to pin 10 (reset pin).

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    value: int, mandatory
        value for pin 10

    Returns
    -------
    True
        if the value is set successfully

    Raises
    ------
    InvalidPin10Value

    Notes
    -----
    N/A

    """
    if value in (0, 1):
        self.PIN_10_SIGNAL_TEST = value
        return True
    raise InvalidPin10Value('PIN 10 attempted to be set to ' + str(value))



"""WPM methods."""


def read_wpm_counter(self) -> str:
    """
    Return the value of the WPM counter ("LEFT" or "RIGHT").

    Internally keeps track of which 4-bits of a register will
    be transferred with the WPM instruction

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    WPM_COUNTER
        The value of the WPM counter

    """
    return self.WPM_COUNTER


def flip_wpm_counter(self) -> str:
    """
    Flip the WPM counter.

    Two WPM instructions must always appear in close succession; that is,
    each time one WPM instruction references a half byte of program RAM
    as indicated by an SRC address, another WPM must access the other half
    byte before the SRC address is altered.
    This internal counter keeps track of which half-byte is being accessed.
    If only one WPM occurs, this counter will be out of sync with the
    program and errors will occur.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    self.WPM_COUNTER
        The flipped value of the WPM counter(either "LEFT" or "RIGHT")

    Raises
    ------
    N/A

    Notes
    -----
    N/A

    """
    if self.WPM_COUNTER == 'LEFT':
        self.WPM_COUNTER = 'RIGHT'
    else:
        self.WPM_COUNTER = 'LEFT'
    return self.WPM_COUNTER



#  Sub-operation methods

from .exceptions import AddressOutOf8BitRange, \
    IncompatibleChunkBit, \
    InvalidBitValue, InvalidChunkValue, \
    InvalidCommandRegisterContent, InvalidCommandRegisterFormat, \
    InvalidPin10Value, InvalidRegister, \
    InvalidRegisterPair, NotABinaryNumber, \
    ProgramCounterOutOfBounds, ValueOutOfRangeForBits, \
    ValueOutOfRangeForStack, ValueTooLargeForAccumulator, \
    ValueTooLargeForRegister, ValueTooLargeForRegisterPair  # noqa


def rdx(self, character):
    """
    Name:           Read RAM status character X

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    character:
        RAM STATUS CHARACTER to read

    Returns
    -------
    self.ACCUMULATOR
        The value read from the specified RAM STATUS CHARACTER

    """
    crb = self.read_current_ram_bank()
    chip, register, _none = \
        decode_command_register(self.COMMAND_REGISTER, 'DATA_RAM_STATUS_CHAR')
    self.ACCUMULATOR = self.STATUS_CHARACTERS[crb][chip][register][character]
    self.increment_pc(1)
    return self.ACCUMULATOR


def encode_command_register(chip, register, address, shape):
    if shape not in ('DATA_RAM_CHAR', 'DATA_RAM_STATUS_CHAR',
                     'RAM_PORT', 'ROM_PORT'):
        raise InvalidCommandRegisterFormat('Shape: ' + shape)

    if shape == 'DATA_RAM_CHAR':
        i_chip = decimal_to_binary(2, chip)
        i_register = decimal_to_binary(2, register)
        i_address = decimal_to_binary(4, address)
        command_register = i_chip + i_register + i_address

    if shape == 'DATA_RAM_STATUS_CHAR':
        i_chip = decimal_to_binary(2, chip)
        i_register = decimal_to_binary(2, register)
        i_address = '0000'
        command_register = i_chip + i_register + i_address

    if shape == 'RAM_PORT':
        # Note that in this instance, "chip" refers to "port"
        i_chip = decimal_to_binary(2, chip)
        i_register = '000'
        i_address = '000'
        command_register = i_chip + i_register + i_address

    if shape == 'ROM_PORT':
        # Note that in this instance, "chip" refers to "port"
        i_chip = decimal_to_binary(4, chip)
        i_register = '00'
        i_address = '00'
        command_register = i_chip + i_register + i_address

    return command_register


def decode_command_register(command_register, shape):
    if shape not in ('DATA_RAM_CHAR', 'DATA_RAM_STATUS_CHAR',
                     'RAM_PORT', 'ROM_PORT'):
        raise InvalidCommandRegisterFormat('Shape: ' + shape)

    command_register = str(command_register)
    if shape == 'DATA_RAM_CHAR':
        if command_register == '0':
            raise InvalidCommandRegisterContent('Content: ' + command_register)
        chip = binary_to_decimal(command_register[:2])
        register = binary_to_decimal(command_register[2:4])
        address = binary_to_decimal(command_register[4:])

    if shape == 'DATA_RAM_STATUS_CHAR':
        if command_register == '0':
            raise InvalidCommandRegisterContent('Content: ' + command_register)
        chip = binary_to_decimal(command_register[:2])
        register = binary_to_decimal(command_register[2:4])
        address = '0'

    if shape == 'RAM_PORT':
        if command_register == '0':
            raise InvalidCommandRegisterContent('Content: ' + command_register)
        # Note that in this instance, "chip" refers to "port"
        chip = binary_to_decimal(command_register[:2])
        register = '0'
        address = '0'

    if shape == 'ROM_PORT':
        if command_register == '0':
            raise InvalidCommandRegisterContent('Content: ' + command_register)
        # Note that in this instance, "chip" refers to "port"
        chip = binary_to_decimal(command_register[:4])
        register = '0'
        address = '0'

    return int(chip), int(register), int(address)


def convert_to_absolute_address(self, rambank, chip, register, address):
    """
    Convert a rambank, chip, register and address to an absolute 0 - 4095
    RAM adddress

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    rambank: integer, mandatory
        The currently selected RAM bank

    chip : integer, mandatory
        1 of 4 chips

    register : integer, mandatory
        1 of 4 registers

    address : integer, mandatory
        address within a page

    Returns
    -------
    absolute_address
        The address from 0 - 4095
    """
    absolute_address = (rambank * self.RAM_BANK_SIZE) + \
        (chip * self.RAM_CHIP_SIZE) + \
        (register * self.RAM_REGISTER_SIZE) + address
    return absolute_address


def split_address8(address):
    """
    Split a supplied decimal address into 2 4-bit words

    Parameters
    ----------
    address: int, mandatory
        An 8-bit address in decimal format

    Returns
    -------
    address_left: str
        left-most 4 bits

    address_right: str
        right-most 4 bits

    Raises
    ------
    N/A

    Notes
    ------
    N/A

    """
    if (address < 0) or (address > 255):
        raise AddressOutOf8BitRange('Address: ' + str(address))

    address_left = bin(address)[2:].zfill(8)[:4]
    address_right = bin(address)[2:].zfill(8)[4:]
    return address_left, address_right


def set_carry(self):
    """
    Set the carry bit

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    self.CARRY
        The carry bit

    Raises
    ------
    N/A

    Notes
    ------
    N/A

    """
    # Set the carry bit
    self.CARRY = 1
    return self.CARRY


def reset_carry(self):
    """
    Resets the carry bit

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    self.CARRY
        The carry bit

    Raises
    ------
    N/A

    Notes
    ------
    N/A

    """
    # Reset the carry bit
    self.CARRY = 0
    return self.CARRY


def read_complement_carry(self):
    """
    Reads the complement of the carry bit, but doesn't change the value

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
        The complement of the carry bit

    Raises
    ------
    N/A

    Notes
    ------
    N/A

    """
    return 1 if self.CARRY == 0 else 0


def insert_register(self, register: int, value: int):
    """
    Insert a value into a specific register

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    register: int, mandatory
        The number of the register to use

    value: int, mandatory
        The value to insert

    Returns
    -------
    value
        The value of the register

    Raises
    ------
    InvalidRegister
    ValueTooLargeForRegister

    Notes
    ------
    N/A

    """
    if (register < 0) or (register > 15):
        raise InvalidRegister('Register: ' + str(register))

    if value > 15:
        raise ValueTooLargeForRegister('Register: ' + str(register) + ',Value: ' + str(value))  # noqa
    self.REGISTERS[register] = value
    return value


def read_register(self, register: int):
    """
    Read a specific register

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    register: int, mandatory
        The number of the register to read

    Returns
    -------
    self.REGISTERS[register]
        The value of the requested register

    Raises
    ------
    InvalidRegister

    Notes
    ------
    N/A

    """
    if (register < 0) or (register > 15):
        raise InvalidRegister('Register:' + str(register))

    return self.REGISTERS[register]


def insert_registerpair(self, registerpair: int, value: int):
    """
    Insert a value into a specific register

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    registerpair: int, mandatory
        The number of the register to insert

    value: int, mandatory
        The value to insert

    Returns
    -------
    value
        The value of the register pair

    Raises
    ------
    InvalidRegisterPair
    ValueTooLargeForRegisterPair

    Notes
    ------
    N/A

    """
    if 0 < registerpair > 7:
        raise InvalidRegisterPair('Register Pair: ' +
                                  str(registerpair))
    if value > 256:
        raise ValueTooLargeForRegisterPair('Register Pair: ' +
                                           str(registerpair) +
                                           ',Value: ' +
                                           str(value))
    # Convert a register pair into a base register for insertion
    base_register = registerpair * 2
    self.insert_register(base_register, (value >> 4) & 15)   # Bit-shift right and remove low bits   # noqa
    self.insert_register(base_register + 1, value & 15)      # Remove low bits                       # noqa
    return value


def read_registerpair(self, registerpair: int):
    """
    Read a specific register pair

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    registerpair: int, mandatory
        The number of the register pair to read

    Returns
    -------
    value
        The value of the requested register pair

    Raises
    ------
    InvalidRegister

    Notes
    ------
    N/A

    """
    if (registerpair < 0 or registerpair > 7):
        raise InvalidRegisterPair('Register Pair: ' +
                                  str(registerpair))
    # Convert a register pair into a base register for insertion
    base_register = registerpair * 2

    hi = self.read_register(base_register)       # High 4-bits
    lo = self.read_register(base_register + 1)   # Low 4-bits
    return (hi << 4) + lo   # Bit-shift left high value and add low value


def increment_pc(self, words: int):
    """
    Increment the Program Counter by a specific number of words

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    words: int, mandatory
        The number of words to increment the Program Counter by

    Returns
    -------
    self.PROGRAM_COUNTER
        The new value of the Program Counter

    Raises
    ------
    ProgramCounterOutOfBounds

    Notes
    ------
    N/A

    """
    if self.PROGRAM_COUNTER + words > self.MEMORY_SIZE_RAM:
        raise ProgramCounterOutOfBounds('Program counter attempted to be' +
                                        ' set to ' +
                                        str(self.PROGRAM_COUNTER + words))
    self.PROGRAM_COUNTER = self.PROGRAM_COUNTER + words
    return self.PROGRAM_COUNTER


def inc_pc_by_page(self, pc: int):
    """
    Retrieve the Program Counter's new value after being incremented
    by a page

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    pc: int, mandatory
        Current value of Program Counter

    Returns
    -------
    pc
        The new value of the Program Counter

    Raises
    ------
    ProgramCounterOutOfBounds

    Notes
    ------
    This function DOES NOT MODIFY the program counter, simply
    calculates the new  value of the counter. It is up to the
    calling  function to determine what to  do with the value.

    """
    if pc + self.PAGE_SIZE > self.MEMORY_SIZE_RAM:
        raise ProgramCounterOutOfBounds('Program counter attempted to be' +
                                        ' set to ' + str(pc + self.PAGE_SIZE))
    # Point the program counter to 1 page on
    pc = pc + self.PAGE_SIZE - 1
    return pc


def is_end_of_page(self, address: int, word: int):
    """
    Determine if an instruction is located at the end of a memory page

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    address: int, mandatory
        Base address of the instruction

    word: int, mandatory
        Number of words in the instruction

    Returns
    -------
    True if the instruction is at the end of the memory page
    False if the instruction is not at the end of the memory page

    Raises
    ------
    N/A

    Notes
    ------
    N/A

    """
    page = address // self.PAGE_SIZE
    location = address - (page * self.PAGE_SIZE)
    word = word - 1
    return (location - word) == self.PAGE_SIZE - 1


def increment_register(self, register: int):
    """
    Increment the value in a register by 1

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    register: int, mandatory
        register to increment

    Returns
    -------
    self.REGISTERS[register]
        value of the register post increment

    Raises
    ------
    InvalidRegister

    Notes
    ------
    N/A

    """
    if register < 0 or register > (self.NO_REGISTERS - 1):
        raise InvalidRegister('Register: ' + str(register))

    self.REGISTERS[register] = self.REGISTERS[register] + 1
    if self.REGISTERS[register] > self.MAX_4_BITS:
        self.REGISTERS[register] = 0
    return self.REGISTERS[register]


def write_pin10(self, value: int):
    """
    Write to pin 10 (reset pin)

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    value: int, mandatory
        value for pin 10

    Returns
    -------
    True
        if the value is set successfully

    Raises
    ------
    InvalidPin10Value

    Notes
    ------
    N/A

    """
    if value in (0, 1):
        self.PIN_10_SIGNAL_TEST = value
        return True
    raise InvalidPin10Value('PIN 10 attempted to be set to ' + str(value))


def write_ram_status(self, char: int):
    """
    Write to a RAM status character

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    char: int, mandatory
        specified status character

    Returns
    -------
    True
        if the value is set successfully

    Raises
    ------
    N/A

    Notes
    -----
    No error checking is done in this function
    All parameters cannot be out of range, since the functions to
    place them in various registers etc all have range checking built in .

    Eventually - there will be error checking here

    """
    value = self.read_accumulator()
    crb = self.read_current_ram_bank()

    chip, register, _none = \
        decode_command_register(self.COMMAND_REGISTER,
                                'DATA_RAM_STATUS_CHAR')
    self.STATUS_CHARACTERS[crb][chip][register][char] = value
    return True

# Miscellaneous read/write operations


def write_to_stack(self, value: int):
    """
    Write to the stack

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    value: int, mandatory
        value to write to the stack

    Returns
    -------
    True
        if the value is written to the stack successfully

    Raises
    ------
    ValueOutOfRangeForStack

    Notes
    ------
    The stack pointer begins at 2, and then moves toward 0

      After 2 writes         After 3 writes         After 3 writes
      +------------+ +------------+ +------------+
      |     a | |      a | <--SP | d |
      |     b | |      b | |      b | <---SP
      | | <--SP | c | |      c |
      +------------+ +------------+ +------------+

    After 3 writes, address "a" is lost
    """
    if (value < 0 or value > 4095):
        raise ValueOutOfRangeForStack(' Value: ' + str(value))

    self.STACK[self.STACK_POINTER] = value
    self.STACK_POINTER = self.STACK_POINTER - 1
    if self.STACK_POINTER == -1:
        self.STACK_POINTER = 2
    return True


def read_from_stack(self):
    """
    Read from the stack

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    value
        the 12-bit number read from the stack

    Raises
    ------
    N/A

    Notes
    ------
    The stack pointer begins at 2, and then moves toward 0

        First Read             Second Read          Third Read
        +------------+ +------------+ +------------+
        |     d | <--SP | d | |      d |
        |     b | |      b | |      b | <---SP
        |     c | |      c | <--SP | c |
        +------------+ +------------+ +------------+

    """
    if self.STACK_POINTER == 2:
        self.STACK_POINTER = 0
    else:
        self.STACK_POINTER = self.STACK_POINTER + 1
    value = self.STACK[self.STACK_POINTER]

    return value

# Utility operations


def ones_complement(value: str, bits: int):
    """
    Converts a decimal value into its one's compliment value
    of a specified bit length

    Parameters
    ----------
    value: int: mandatory
        decimal value to convert

    bits: int, mandatory
        number of bits required for the conversion

    Returns
    -------
    The one's compliment binary value of the supplied decimal value.

    Raises
    ------
    InvalidBitValue: When a bit value of not 4, 8 or 12 is specified
    ValueOutOfRangeForBits: If the value supplied is either negative or is
                             out of range of the number of bits requested

    Notes
    ------
    N/A

    """
    if (bits not in [2, 4, 8, 12]):
        raise InvalidBitValue(' Bits: ' + str(bits))

    if (value > ((2 ** bits) - 1)) or (value < 0):
        raise ValueOutOfRangeForBits(' Value: ' + str(value) +
                                     ' Bits: ' + str(bits))

    # Perform a one's complement
    # i.e. invert all the bits

    binary = bin(value)[2:].zfill(bits)
    ones = ''
    for x in range(bits):
        if binary[x] == '1':
            ones = ones + '0'
        else:
            ones = ones + '1'
    return ones


def convert_decimal_to_n_bit_slices(bits: int, chunk: int, decimal: int, result: str = 'b'):  # noqa
    """
    Converts a decimal value into several binary or decimal values of specific
    bit lengths

    Parameters
    ----------
    bits: int, mandatory
        number of bits of the source data

    chunk: int, mandatory
        number of bits required per chunk

    decimal: int: mandatory
        decimal value to convert

    result: str: mandatory
        'd' will generate a decimal output
        'b' will generate a binary output

    Returns
    -------
    The binary value of the supplied decimal value.

    Raises
    ------
    IncompatibleChunkBit: When the chunks do not fit exactly within the bits
    InvalidBitValue: When a bit value of not 4, 8 or 12 is specified
    InvalidChunkValue: When a chunk value of not 4, 8 or 12 is specified
    ValueOutOfRangeForBits: If the value supplied is either negative or is
                             out of range of the number of bits requested

    Notes
    ------
    N/A

    """
    if (bits not in [2, 4, 8, 12]):
        raise InvalidBitValue(' Bits: ' + str(bits))

    if (chunk not in [2, 4, 8, 12]):
        raise InvalidChunkValue(' Chunk: ' + str(chunk))

    if bits % chunk != 0:
        raise IncompatibleChunkBit(' Bits: ' + str(bits) +
                                   ' Chunk: ' + str(chunk))

    if (decimal > ((2 ** bits) - 1)) or (decimal < 0):
        raise ValueOutOfRangeForBits(' Value: ' + str(decimal) +
                                     ' Bits: ' + str(bits))

    binary = decimal_to_binary(bits, decimal)
    chunks = [binary[i:i+chunk] for i in range(0, len(binary), chunk)]
    if result != 'b':
        decimals = []
        for element in chunks:
            decimals.append(binary_to_decimal(element))
        chunks = decimals
    return chunks


def decimal_to_binary(bits: int, decimal: int):
    """
    Converts a decimal value into a binary value of a specified bit length

    Parameters
    ----------
    bits: int, mandatory
        number of bits required for the conversion

    decimal: int: mandatory
        decimal value to convert

    Returns
    -------
    The binary value of the supplied decimal value.

    Raises
    ------
    InvalidBitValue: When a bit value of not 4, 8 or 12 is specified
    ValueOutOfRangeForBits: If the value supplied is either negative or is
                             out of range of the number of bits requested

    Notes
    ------
    N/A

    """
    if (bits not in [2, 4, 8, 12]):
        raise InvalidBitValue(' Bits: ' + str(bits))

    if (decimal > ((2 ** bits) - 1)) or (decimal < 0):
        raise ValueOutOfRangeForBits(' Value: ' + str(decimal) +
                                     ' Bits: ' + str(bits))

    # Convert decimal to binary
    binary = bin(decimal)[2:].zfill(bits)
    return binary


def binary_to_decimal(binary: str):
    """
    Converts a string value(which must be in binary form) to
    a decimal value

    Parameters
    ----------
    binary: str, mandatory
        a string which represents the binary value

    Returns
    -------
    The decimal value of the supplied binary value

    Raises
    ------
    NotABinaryNumber: When a non-binary number is supplied

    Notes
    ------
    N/A

    """
    if len(binary) == 0:
        binary = '<empty>'
        raise NotABinaryNumber('"' + binary + '"')

    if len(binary.replace('0', '').replace('1', '')) != 0:
        raise NotABinaryNumber('"' + binary + '"')

    # Convert binary to decimal
    return int(binary, 2)


def flip_wpm_counter(self):
    """
    Two WPM instructions must always appear in close succession; that is ,
    each time one WPM instruction references a half byte of program RAM
    as indicated by an SRC address, another WPM must access the other half
    byte before the SRC address is altered.
    This internal counter keeps track of which half-byte is being accessed.
    If only one WPM occurs, this counter will be out of sync with the
    program and errors will occur.

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    self.WPM_COUNTER
        The flipped value of the WPM counter(either "LEFT" or "RIGHT")

    Raises
    ------
    N/A

    Notes
    ------
    N/A

    """
    if self.WPM_COUNTER == 'LEFT':
        self.WPM_COUNTER = 'RIGHT'
    else:
        self.WPM_COUNTER = 'LEFT'
    return self.WPM_COUNTER


def check_overflow(self):
    """
    Check for an overflow is detected
    i.e. the result is more than a 4-bit number(MAX_4_BITS)

    If there is an overflow detected, set the carry bit,
    otherwise reset the carry bit.

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    self.ACCUMULATOR
        The value of the accumulator after adjusting for overflow
    self.CARRY
        The carry bit

    Raises
    ------
    N/A

    Notes
    ------
    N/A

    """
    if self.ACCUMULATOR > self.MAX_4_BITS:
        self.ACCUMULATOR = self.ACCUMULATOR - self.MAX_4_BITS + 1
        self.set_carry()
    else:
        self.reset_carry()
    return self.ACCUMULATOR, self.CARRY


def set_accumulator(self, value: int):
    """
    Insert a value into the Accumulator

    Parameters
    ----------
    self: processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    value: int, mandatory
        The value to insert

    Returns
    -------
    value
        The value of the accumulator

    Raises
    ------
    ValueTooLargeForAccumulator

    Notes
    ------
    N/A

    """
    if value > self.MAX_4_BITS:
        raise ValueTooLargeForAccumulator(' Value: ' + str(value))
    self.ACCUMULATOR = value
    return value



"""Initialisation methods."""


def init_ram(self) -> None:
    """
    Initialise the RAM with zeroes in all locations.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    """
    for _i in range(self.MEMORY_SIZE_RAM):
        self.RAM.append(0)


def init_command_registers(self) -> None:
    """
    Initialise the command registers with zeroes.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    """
    for _i in range(self.NO_COMMAND_REGISTERS):
        self.COMMAND_REGISTERS.append(0)


def init_registers(self) -> None:
    """
    Initialise the registers with zeroes.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    """
    for _i in range(self.NO_REGISTERS):
        self.REGISTERS.append(0)


def init_stack(self) -> None:
    """
    Initialise the stack.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    """
    for _i in range(self.STACK_SIZE):
        self.STACK.append(0)


def init_rom(self) -> None:
    """
    Initialise the ROM with zeroes in all locations.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    """
    for _i in range(self.MEMORY_SIZE_ROM):
        self.ROM.append(0)


def init_pram(self) -> None:
    """
    Initialise the PRAM with zeroes in all locations.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    """
    for _i in range(self.MEMORY_SIZE_PRAM):
        self.PRAM.append(0)


def init_wpm_counter(self) -> None:
    """
    Initialise the WPM Counter.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Notes
    -----
    The WPM counter is required to allow WPM instructions to track which
    4-bit portion of an 8-bit byte is being transferred.

    """
    self.WPM_COUNTER = 'LEFT'



"""ROM methods."""


def read_all_rom(self) -> list:
    """
    Return the values of all the locations of ROM.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    ROM
        The values of all the locations of ROM

    """
    return self.ROM


def read_all_rom_ports(self) -> list:
    """
    Return the values of all the ROM ports.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    ROM_PORT
        The values of all the ROM ports

    """
    return self.ROM_PORT



"""Exceptions to raise when issues occur."""


class AddressOutOf8BitRange(Exception):

    """Raised when an address is out of range of an 8-bit Value."""


class IncompatibleChunkBit(Exception):

    """Raised when a bit value is not a multiple of the chunk value."""


class InvalidBitValue(Exception):

    """Raised when a bit value of other than 4,8 or 12 is used."""


class InvalidChunkValue(Exception):

    """Raised when a chunk value of other than 4,8 or 12 is used."""


class InvalidCommandRegisterContent(Exception):
    """
    Raised when a command register contains an illegally formatted value.
    """


class InvalidCommandRegisterFormat(Exception):

    """Raised when a CR format isn't within the prescribed values."""


class InvalidEndOfPage(Exception):

    """Raised when it is impossible to determine the end of a page."""


class InvalidPin10Value(Exception):

    """Raised when the value of PIN 10 is attempted to be set to NOT 0 or 1."""


class InvalidRamBank(Exception):

    """Raised when the attempting to select a RAM bank > 7."""


class InvalidRegister(Exception):

    """Raised when an invalid register is supplied."""


class InvalidRegisterPair(Exception):

    """Raised when an invalid register pair is supplied."""


class NotABinaryNumber(Exception):

    """Raised when a supplied binary number is NOT binary."""


class ProgramCounterOutOfBounds(Exception):

    """Raised when the program counter is forced beyond the end of memory."""


class ValueTooLargeForAccumulator(Exception):

    """Raised when the value for the Accumulator is too large."""


class ValueOutOfRangeForBits(Exception):

    """Raised when the value is larger than the number of bits supplied."""


class ValueTooLargeForRegister(Exception):

    """Raised when the value for a register is too large."""


class ValueTooLargeForRegisterPair(Exception):

    """Raised when the value for a register pair is too large."""


class ValueOutOfRangeForStack(Exception):

    """Raised when the value for the Stack is out of Range."""



"""Read Processor methods."""

# Import typing library
from typing import Tuple

from hardware.exceptions import InvalidCommandRegisterContent, InvalidCommandRegisterFormat  # noqa
from hardware.suboperations.utility import binary_to_decimal


def read_all_command_registers(self) -> list:
    """
    Return the values of all of the Command Registers.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    COMMAND_REGISTERS
        The values of all the Command Registers

    """
    return self.COMMAND_REGISTERS


def decode_command_register(command_register: str,
                            shape: str) -> Tuple[int, int, int]:
    """
    Convert the supplied CR into its component parts.

    Parameters
    ----------
    command_register : str, mandatory
        Content of the command register to convert

    shape:
        The shape/purpose of the command_register

    Returns
    -------
    chip: int
        The chip referred to

    register: int
        register

    address: int
        address referred to

    Raises
    ------
    InvalidCommandRegisterFormat

    """
    if shape not in ('DATA_RAM_CHAR', 'DATA_RAM_STATUS_CHAR',
                     'RAM_PORT', 'ROM_PORT'):
        raise InvalidCommandRegisterFormat('Shape: ' + shape)

    command_register = str(command_register)
    if shape == 'DATA_RAM_CHAR':
        if command_register == '0':
            raise InvalidCommandRegisterContent('Content: ' + command_register)
        chip = binary_to_decimal(command_register[:2])
        register = binary_to_decimal(command_register[2:4])
        address = binary_to_decimal(command_register[4:])

    if shape == 'DATA_RAM_STATUS_CHAR':
        if command_register == '0':
            raise InvalidCommandRegisterContent('Content: ' + command_register)
        chip = binary_to_decimal(command_register[:2])
        register = binary_to_decimal(command_register[2:4])
        address = '0'

    if shape == 'RAM_PORT':
        if command_register == '0':
            raise InvalidCommandRegisterContent('Content: ' + command_register)
        # Note that in this instance, "chip" refers to "port"
        chip = binary_to_decimal(command_register[:2])
        register = '0'
        address = '0'

    if shape == 'ROM_PORT':
        if command_register == '0':
            raise InvalidCommandRegisterContent('Content: ' + command_register)
        # Note that in this instance, "chip" refers to "port"
        chip = binary_to_decimal(command_register[:4])
        register = '0'
        address = '0'

    return int(chip), int(register), int(address)



###########################################################
#  _ _  _    ___   ___  _  _     _           _            #
# (_) || |  / _ \ / _ \| || |   (_)         | |           #
#  _| || |_| | | | | | | || |_   _ _ __   __| | _____  __ #
# | |__   _| | | | | | |__   _| | | '_ \ / _` |/ _ \ \/ / #
# | |  | | | |_| | |_| |  | |   | | | | | (_| |  __/>  <  #
# |_|  |_|  \___/ \___/   |_|   |_|_| |_|\__,_|\___/_/\_\ #
#                                                         #
###########################################################

"""
Commands in this module.

            INC -   INCREMENT REGISTER
            FIN -   FETCH INDIRECT


 Abbreviations used in the descriptions of each instruction's actions:

            (    )      the content of
            -->         is transferred to
            ACC	        Accumulator (4-bit)
            CY	        Carry/link Flip-Flop
            ACBR	    Accumulator Buffer Register (4-bit)
            RRRR	    Index register address
            RPn	        Index register pair address
            PL	        Low order program counter Field (4-bit)
            PM	        Middle order program counter Field (4-bit)
            PH	        High order program counter Field (4-bit)
            ai	        Order i content of the accumulator
            CMi	        Order i content of the command register
            M	        RAM main character location
            MSi	        RAM status character i
            DB (T)	    Data bus content at time T
            Stack	    The 3 registers in the address register
                        other than the program counter

    Additional Abbreviations:
            ~           Inverse (1's complement)
            .           logical OR

"""

# Import typing library
from typing import Tuple


def fin(self, registerpair: int) -> Tuple[int, int]:
    """
    Name:           Fetch indirect from ROM.

    Function:       The 8 bit content of the 0 index register
                    pair (0000) (0001) is sent out as an address
                    in the same page where the FIN instruction is located.
                    The 8 bit word at that location is loaded into the
                    designated index register pair. The program counter
                    is unaffected; after FIN has been executed the next
                    instruction in sequence will be addressed. The content
                    of the 0 index register pair is unaltered unless
                    index register 0 was designated.
    Syntax:         FIN
    Assembled:      0011 RRRO
    Symbolic:       (PH) (0000) (0001) --> ROM address
                    (OPR) --> RRRO
                    (OPA) --> RRR1
    Execution:      1 word, 16-bit code and an execution time of 21.6 usec.
    Side-effects:   Not Applicable
    Exceptions:     a) Although FIN is a 1-word instruction, its execution
                        requires two memory cycles (21.6 psec).
                    b) When FIN is located at address (PH) 1111 1111 data will
                        be fetched from the next page(ROM) in sequence and not
                        from the same page(ROM) where the FIN instruction is
                        located. That is, next address is
                        (PH + 1) (0000) (0001) and not (PH) (0000) (0001).
    """
    eop = self.is_end_of_page(self.PROGRAM_COUNTER, 1)
    if eop is True:
        page_shift = 1
    else:
        page_shift = 0
    value = self.RAM[(self.REGISTERS[1] + (self.REGISTERS[0] << 4)) +
                     (self.PAGE_SIZE * page_shift)]
    self.insert_registerpair(registerpair, value)
    self.increment_pc(1)
    return self.REGISTERS[registerpair], self.REGISTERS[registerpair+1]


def inc(self, register: int) -> int:
    """
    Name:           Increment index register.

    Function:       The 4 bit content of the designated index register is
                    incremented by 1.
                    The index register is set to zero in case of overflow.
    Syntax:         INC <register>
    Assembled:      0110 <RRRR>
    Symbolic:       (RRRR) +1 --> RRRR
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry bit is not affected.
    """
    self.increment_register(register)
    self.increment_pc(1)
    return self.REGISTERS[register]



##########################################################################
#                         _ _ _   __   __  _ _                           #
#                        (_) | | /  \ /  \| | |                          #
#                        | |_  _| () | () |_  _|                         #
#                        |_| |_| \__/ \__/  |_|                          #
#       _                     __                      _           _      #
#      | |_ _ _ __ _ _ _  ___/ _|___ _ _   __ ___ _ _| |_ _ _ ___| |     #
#      |  _| '_/ _` | ' \(_-<  _/ -_) '_| / _/ _ \ ' \  _| '_/ _ \ |     #
#       \__|_| \__,_|_||_/__/_| \___|_|   \__\___/_||_\__|_| \___/_|     #
#                                                                        #
##########################################################################

"""
Commands in this module.

            JUN -   JUMP UNCONDITIONALLY
            JIN -   JUMP INDIRECT
            JCN -   JUMP ON CONDITION
            ISZ -   INCREMENT AND SKIP IF ZERO

 Abbreviations used in the descriptions of each instruction's actions:

            (    )      the content of
            -->         is transferred to
            ACC	        Accumulator (4-bit)
            CY	        Carry/link Flip-Flop
            ACBR	    Accumulator Buffer Register (4-bit)
            RRRR	    Index register address
            RPn	        Index register pair address
            PL	        Low order program counter Field (4-bit)
            PM	        Middle order program counter Field (4-bit)
            PH	        High order program counter Field (4-bit)
            ai	        Order i content of the accumulator
            CMi	        Order i content of the command register
            M	        RAM main character location
            MSi	        RAM status character i
            DB (T)	    Data bus content at time T
            Stack	    The 3 registers in the address register
                        other than the program counter

    Additional Abbreviations:
            ~           Inverse (1's complement)
            .           logical OR

"""


def jun(self, address: int) -> int:
    """
    Name:           Jump unconditional.

    Function:       Program control is unconditionally
                    transferred to the instruction located
                    at the address AAAA3, AAAA2, AAAA1.
    Syntax:         JUN
    Assembled:      0100 AAAA3
                    AAAA2 AAAA1
    Symbolic:       AAAA1 --> PL,
                    AAAA2 --> PM,
                    AAAA3 --> PH
    Execution:      2 words, 16-bit code and an execution time of 21.6 usec.
    Side-effects:   Not Applicable
    """
    from hardware.exceptions import ProgramCounterOutOfBounds  # noqa

    if (address >= self.MEMORY_SIZE_RAM or address < 0):
        raise ProgramCounterOutOfBounds('Program counter attempted to be' +
                                        ' set to ' +
                                        str(address))
    self.PROGRAM_COUNTER = address
    return self.PROGRAM_COUNTER


def jin(self, registerpair: int) -> int:
    """
    Name:           Jump Indirect.

    Function:       The 8 bit content of the designated index register pair
                    is loaded into the low order 8 positions of the program
                    counter.
                    Program control is transferred to the instruction at
                    that address on the same page (same ROM) where the JIN
                    instruction is located.
                    The 8 bit content of the index register is unaffected.
    Syntax:         JIN
    Assembled:      0011 RRR1
    Symbolic:       (RRRO) --> PM
                    (RRR1) --> PL
                    PH unchanged
    Execution:      1 words, 16-bit code and an execution time of 10.8 usec.
    """
    from hardware.exceptions import ProgramCounterOutOfBounds  # noqa

    address = self.read_registerpair(registerpair)
    pcb = self.PROGRAM_COUNTER

    # Increment PROGRAM_COUNTER by a page if the instruction is at
    # the last position in a page.
    if self.is_end_of_page(self.PROGRAM_COUNTER, 1) is True:
        self.PROGRAM_COUNTER = self.inc_pc_by_page(self.PROGRAM_COUNTER - 1)
        self.PROGRAM_COUNTER = self.PROGRAM_COUNTER - 12
        address = address - 1

    self.PROGRAM_COUNTER = self.PROGRAM_COUNTER >> 8
    self.PROGRAM_COUNTER = (self.PROGRAM_COUNTER << 8) + address

    if self.PROGRAM_COUNTER >= self.MEMORY_SIZE_RAM:
        e_pc = self.PROGRAM_COUNTER
        self.PROGRAM_COUNTER = pcb
        raise ProgramCounterOutOfBounds('Program counter attempted to be' +
                                        ' set to ' +
                                        str(e_pc))

    return self.PROGRAM_COUNTER


def jcn(self, conditions: int, address: int) -> int:
    """
    Name:           Jump conditional.

    Function:       If the designated condition code is true, program control
                    is transferred to the instruction located at the 8 bit
                    address AAAA2, AAAA1 on the same page (ROM) where JCN is
                    located.
                    If the condition is not true the next instruction in
                    sequence after JCN is executed.
                    The condition bits are assigned as follows:
                    C1 = 0 Do not invert jump condition }
                    C1 = 1 Invert jump condition        }
                    C2 = 1 Jump if the accumulator content is zero
                    C3 = 1 Jump if the carry/link content is 1
                    C4 = 1 Jump if test signal (pin 10 on 4004) is zero.
    Syntax:         JCN
    Assembled:      0001 C1C2C3C4
                    AAAA2 AAAA1
    Symbolic:       If C1C2C3C4 is true, A2A2A2A2 --> PM
                    A1A1A1A1 --> PL, PH unchanged
                    if C1C2C3C4 is false,
                    (PH) --> PH, (PM) --> PM, (PL + 2) --> PL
    Execution:      2 words, 16-bit code and an execution time of 21.6 usec.

    Sample:

    OPR     OPA
    0001    0110  Jump if accumulator is zero or carry = 1

    Several conditions can be tested simultaneously.

    The logic equation describing the condition for a jump is give below:
    JUMP = ~C1 . ((ACC = 0) . C2 + (CY = 1) . C3 + ~TEST . C4) +
                C1 . ~((ACC != 0) . C2 + (CY = 1) . C3 + ~TEST . C4)

                        +---------+---------+
                        | Symbol  | Logical |
                        +---------+---------+
                        |    ~    +   NOT   +
                        |    .    +   AND   +
                        |    +    +   OR    +
                        +---------+---------+

    Side-effects:   Not Applicable

    Info about signal/test pin 10 on intel4004
    https://tams.informatik.uni-hamburg.de/applets/hades/webdemos/80-mcs4/jmp/jmp_test.html

    Assembler:
            jcn     IACT    lbl

    I - Invert other conditions
    A - Accumulator = 0
    C - Carry Bit set (i.e. = 1)
    T - Test Signal on Intel4004 Pin 10 = 0

    Need to do "if JCN at end of page" code

    """
    from hardware.suboperations.utility import decimal_to_binary  # noqa

    accumulator = self.read_accumulator()
    checks = str(decimal_to_binary(4, conditions))

    c1 = checks[0:1] == '1'
    c2 = checks[1:2] == '1'
    c3 = checks[2:3] == '1'
    c4 = checks[3:4] == '1'
    notc1 = True if c1 is False else True
    notc2 = True if c2 is False else True
    notc3 = True if c3 is False else True
    notc4 = True if c4 is False else True

    # Use symbolic logic to determine whether to jump
    jump = notc1 and ((accumulator == 0) and c2 or (self.read_carry() == 1)
                      and c3 or (not self.read_pin10()) and c4) or \
        c1 and (((accumulator != 0) or notc2) and
                ((self.read_carry() == 0) or notc3) and
                ((not self.read_pin10()) or notc4))

    if jump is True:
        self.PROGRAM_COUNTER = address
    else:
        self.increment_pc(2)
    return self.PROGRAM_COUNTER


def isz(self, register: int, address: int) -> int:
    """
    Name:           Increment index register skip if zero.

    Function:       The content of the designated index register is
                    incremented by 1.
                    The accumulator and carry/link are unaffected.
                    If the result is zero, the next instruction after ISZ is
                    executed.
                    If the result is different from 0, program control is
                    transferred to the instruction located at the 8 bit
                    address AAAA2, AAAA1 on the same page (ROM) where the
                    ISZ instruction is located.
    Syntax:         ISZ
    Assembled:      0111 RRRR
                    AAAA2 AAAA1
    Symbolic:       (RRRR) + 1 --> RRRR,
                    if result = 0 (PH) --> PH, (PM) --> PM, (PL + 2) --> PL
                    if result <> 0  (PH) --> PH, AAAA2 --> PM, `AAAA1 --> PL
    Execution:      2 words, 16-bit code and an execution time of 21.6 usec.
    Side-effects:   Not Applicable


    """
    self.increment_register(register)
    if self.REGISTERS[register] == 0:
        self.increment_pc(2)
    else:
        self.PROGRAM_COUNTER = address
    return self.PROGRAM_COUNTER



##########################################################################
#                         _ _ _   __   __  _ _                           #
#                        (_) | | /  \ /  \| | |                          #
#                        | |_  _| () | () |_  _|                         #
#                        |_| |_| \__/ \__/  |_|                          #
#                                                                        #
#                     ___ ___    ___    _   __  __                       #
#                    |_ _/ _ \  | _ \  /_\ |  \/  |                      #
#                     | | (_) | |   / / _ \| |\/| |                      #
#                    |___\___/  |_|_\/_/ \_\_|  |_|                      #
#                                                                        #
##########################################################################

"""
Commands in this module.

            RDM -   READ DATA RAM DATA CHARACTER
            RDn -   READ DATA RAM STATUS CHARACTER
            RDR -   READ DATA ROM PORT
            WRM -   WRITE DATA RAM CHARACTER
            WRn -   WRITE DATA RAM STATUS CHARACTER
            WMP -   WRITE RAM PORT
            WRR -   WRITE ROM PORT
            ADM -   ADD DATA RAM TO ACCUMULATOR WITH CARRY
            SBM -   SUBTRACT DATA RAM FROM ACCUMULATOR WITH BORROW
            WPM -   WRITE PROGRAM RAM


 Abbreviations used in the descriptions of each instruction's actions:

            (    )      the content of
            -->         is transferred to
            ACC	        Accumulator (4-bit)
            CY	        Carry/link Flip-Flop
            ACBR	    Accumulator Buffer Register (4-bit)
            RRRR	    Index register address
            RPn	        Index register pair address
            PL	        Low order program counter Field (4-bit)
            PM	        Middle order program counter Field (4-bit)
            PH	        High order program counter Field (4-bit)
            ai	        Order i content of the accumulator
            CMi	        Order i content of the command register
            M	        RAM main character location
            MSi	        RAM status character i
            DB (T)	    Data bus content at time T
            Stack	    The 3 registers in the address register
                        other than the program counter

    Additional Abbreviations:
            ~           Inverse (1's complement)
            .           logical OR

"""

# Import system modules
import os
import sys
sys.path.insert(1, '..' + os.sep + 'src')

# Import typing library
from typing import Tuple  # noqa

from hardware.suboperations.utility import convert_to_absolute_address, \
    decimal_to_binary, ones_complement  # noqa
from hardware.suboperations.other import decode_command_register  # noqa
from hardware.suboperations.accumulator import check_overflow  # noqa
from hardware.suboperations.ram import rdx  # noqa
from hardware.suboperations.wpm import flip_wpm_counter, read_wpm_counter  # noqa


def rdm(self) -> int:
    """
    Name:           Read Data RAM data character.

    Function:       The content of the previously selected RAM main memory
                    character is transferred to the accumulator
                    The carry/link is unaffected.
                    The 4-bit data in memory is unaffected.
    Syntax:         RDM
    Assembled:      1110 1001
    Symbolic:       (M) --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    """
    crb = self.read_current_ram_bank()
    chip, register, address = \
        decode_command_register(self.COMMAND_REGISTER, 'DATA_RAM_CHAR')
    absolute_address = convert_to_absolute_address(
        self, crb, chip, register, address)
    self.ACCUMULATOR = self.RAM[absolute_address]
    self.increment_pc(1)
    return self.ACCUMULATOR


def rd0(self) -> int:
    """
    Name:           Read RAM status character 0.

    Function:       The 4-bits of status character 0 for the previously
                    selected RAM register are transferred to the
                    accumulator.
                    The carry/link and the status character are unaffected.
    Syntax:         RD0
    Assembled:      1110 1100
    Symbolic:       (MS0) --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)
    Bits 1 + 2 = 1 of 4 DATA RAM CHIPS within the DATA RAM BANK previously
                 selected by a DCL instruction
    Bits 3 + 4 = 1 of 4 registers within the DATA RAM CHIP
    Bits 5-8   = Not relevant

    """
    return rdx(self, 0)


def rd1(self) -> int:
    """
    Name:           Read RAM status character 1.

    Function:       The 4-bits of status character 1 for the previously
                    selected RAM register are transferred to the
                    accumulator.
                    The carry/link and the status character are unaffected.
    Syntax:         RD1
    Assembled:      1110 1101
    Symbolic:       (MS1) --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)
    Bits 1 + 2 = 1 of 4 DATA RAM CHIPS within the DATA RAM BANK previously
                 selected by a DCL instruction
    Bits 3 + 4 = 1 of 4 registers within the DATA RAM CHIP
    Bits 5-8   = Not relevant

    """
    return rdx(self, 1)


def rd2(self) -> int:
    """
    Name:           Read RAM status character 2.

    Function:       The 4-bits of status character 2 for the previously
                    selected RAM register are transferred to the
                    accumulator.
                    The carry/link and the status character are unaffected.
    Syntax:         RD2
    Assembled:      1110 1101
    Symbolic:       (MS2) --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)
    Bits 1 + 2 = 1 of 4 DATA RAM CHIPS within the DATA RAM BANK previously
                 selected by a DCL instruction
    Bits 3 + 4 = 1 of 4 registers within the DATA RAM CHIP
    Bits 5-8   = Not relevant

    """
    return rdx(self, 2)


def rd3(self) -> int:
    """
    Name:           Read RAM status character 3.

    Function:       The 4-bits of status character 3 for the previously
                    selected RAM register are transferred to the
                    accumulator.
                    The carry/link and the status character are unaffected.
    Syntax:         RD3
    Assembled:      1110 1111
    Symbolic:       (MS3) --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)
    Bits 1 + 2 = 1 of 4 DATA RAM CHIPS within the DATA RAM BANK previously
                 selected by a DCL instruction
    Bits 3 + 4 = 1 of 4 registers within the DATA RAM CHIP
    Bits 5-8   = Not relevant

    """
    return rdx(self, 3)


def rdr(self) -> int:
    """
    Name:           Read ROM Port.

    Function:       The ROM port specified by the last SRC instruction is read.
                    When using the 4001 ROM,each of the 4 lines of the port may
                    be an input or an output line; the data on the input lines
                    is transferred to the corresponding bits of the
                    accumulator.
                    Any output lines cause either a 0 or a 1 to be transferred
                    to the corresponding bits of the accumulator.
                    Whether a 0 or a 1 is transferred is a function of the
                    hardware, not under control of the programmer. (See Note 3)
                    The Carry bit is unaffected.
    Syntax:         RDR
    Assembled:      1110 1010
    Symbolic:       (ROM input lines) --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.

    Notes:  1       The i4001 chip has a 256-byte ROM (256 8-bit program
                    instructions), and one built-in 4-bit I/O port.
                    The MCS-4 has 16 i4001 ROM chips.

            2       On the INTELLEC 4, a ROM port may be used for either
                    input or output. If programs tested on the INTELLEC 4 are
                    to be run later with a 4001 ROM the programmer must be
                    careful not to use one port for both functions.

            3       On the physical devices, if the leftmost I/O line is an
                    output line and the remaining I/O lines are input lines
                    containing 010B, the accumulator will contain either
                    1010B or 0010B.

    Implementation  This software implementation of the i4004 will ALWAYS
                    return the values of the output lines as-is.
    """
    rom, _unused1, _unused2 = \
        decode_command_register(self.COMMAND_REGISTER, 'ROM_PORT')
    self.ACCUMULATOR = self.ROM_PORT[rom]
    self.increment_pc(1)
    return self.ACCUMULATOR


def wrm(self) -> int:
    """
    Name:           Write accumulator into RAM character.

    Function:       The accumulator content is written into the previously
                    selected RAM main memory character location.
                    The accumulator and carry/link are unaffected.
    Syntax:         WRM
    Assembled:      1110 0000
    Symbolic:       (ACC) --> M
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable
    """
    value = self.ACCUMULATOR
    crb = self.read_current_ram_bank()
    chip, register, address = \
        decode_command_register(self.COMMAND_REGISTER, 'DATA_RAM_CHAR')
    absolute_address = convert_to_absolute_address(
        self, crb, chip, register, address)
    self.RAM[absolute_address] = value
    self.increment_pc(1)
    return self.PROGRAM_COUNTER


def wr0(self) -> int:
    """
    Name:           Write accumulator into RAM status character 0.

    Function:       The content of the accumulator is written into the
                    RAM status character 0 of the previously selected
                    RAM register.
                    The accumulator and the carry/link are unaffected.
    Syntax:         WR0
    Assembled:      1110 0100
    Symbolic:       (ACC) --> MS0
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)
    Bits 1 + 2 = 1 of 4 DATA RAM CHIPS within the DATA RAM BANK previously
                 selected by a DCL instruction
    Bits 3 + 4 = 1 of 4 registers within the DATA RAM CHIP
    Bits 5-8   = Not relevant

    """
    self.write_ram_status(0)
    self.increment_pc(1)
    return self.PROGRAM_COUNTER


def wr1(self) -> int:
    """
    Name:           Write accumulator into RAM status character 1.

    Function:       The content of the accumulator is written into the
                    RAM status character 1 of the previously selected
                    RAM register.
                    The accumulator and the carry/link are unaffected.
    Syntax:         WR1
    Assembled:      1110 0101
    Symbolic:       (ACC) --> MS1
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)
    Bits 1 + 2 = 1 of 4 DATA RAM CHIPS within the DATA RAM BANK previously
                 selected by a DCL instruction
    Bits 3 + 4 = 1 of 4 registers within the DATA RAM CHIP
    Bits 5-8   = Not relevant

    """
    self.write_ram_status(1)
    self.increment_pc(1)
    return self.PROGRAM_COUNTER


def wr2(self) -> int:
    """
    Name:           Write accumulator into RAM status character 2.

    Function:       The content of the accumulator is written into the
                    RAM status character 2 of the previously selected
                    RAM register.
                    The accumulator and the carry/link are unaffected.
    Syntax:         WR2
    Assembled:      1110 0110
    Symbolic:       (ACC) --> MS2
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)
    Bits 1 + 2 = 1 of 4 DATA RAM CHIPS within the DATA RAM BANK previously
                 selected by a DCL instruction
    Bits 3 + 4 = 1 of 4 registers within the DATA RAM CHIP
    Bits 5-8   = Not relevant

    """
    self.write_ram_status(2)
    self.increment_pc(1)
    return self.PROGRAM_COUNTER


def wr3(self) -> int:
    """
    Name:           Write accumulator into RAM status character 3.

    Function:       The content of the accumulator is written into the
                    RAM status character 3 of the previously selected
                    RAM register.
                    The accumulator and the carry/link are unaffected.
    Syntax:         WR3
    Assembled:      1110 0111
    Symbolic:       (ACC) --> MS3
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)
    Bits 1 + 2 = 1 of 4 DATA RAM CHIPS within the DATA RAM BANK previously
                 selected by a DCL instruction
    Bits 3 + 4 = 1 of 4 registers within the DATA RAM CHIP
    Bits 5-8   = Not relevant

    """
    self.write_ram_status(3)
    self.increment_pc(1)
    return self.PROGRAM_COUNTER


def wmp(self) -> int:
    """
    Name:           Write memory port.

    Function:       The content of the accumulator is transferred to the
                    RAM output port of the previously selected RAM chip.
                    The data is available on the output pins until a new
                    WMP is executed on the same RAM chip.
                    The content of the ACC and the carry/link are unaffected.
    Syntax:         WMP
    Assembled:      1110 0001
    Symbolic:       (ACC) --> RAM output register
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The LSB bit of the accumulator appears on I/O 0, Pin 16,
                    of the 4002 RAM chip until it is changed.

    There is a 4-bit register on each DATA RAM Chip -  on the base MCS-4
    system (one i4004 CPU, 32 * i4002, and ? * i4001 ) - additionally, a i4003
    shift register.

    Therefore - 4 chips per bank, 8 banks = 32 addressable memory ports.

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)

    Bits 1 - 2 = The port associated with 1 of 4 DATA RAM
                 chips within the DATA RAM bank previously
                 selected by a DCL instruction
    Bits 3 - 8 = Not relevant
    """
    crb = self.read_current_ram_bank()
    chip, _unused1, _unused2 = \
        decode_command_register(self.COMMAND_REGISTER, 'RAM_PORT')
    self.RAM_PORT[crb][chip] = self.ACCUMULATOR
    self.increment_pc(1)
    return self.ACCUMULATOR


def wrr(self) -> int:
    """
    Name:           Write ROM port.

    Function:       The content of the accumulator is transferred to the ROM
                    output port of the previously selected ROM chip.
                    The data is available on the output pins until a new WRR
                    is executed on the same chip.
                    The content of the ACC and the carry/link are unaffected.
    Syntax:         WRR
    Assembled:      1110 0010
    Symbolic:       (ACC) --> ROM output lines
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The LSB bit of the accumulator appears on I/O 0, Pin 16,
                    of the 4001 ROM chip until it is changed.
    Notes:          No operation is performed on I/O lines coded as inputs.

    4 chips per bank, 8 banks = 32 addressable ROM ports.

    An address set by the previous SRC instruction is interpreted as follows:

    (Bits in this order : 12345678)

    Bits 1 - 4 = The ROM chip targetted
    Bits 5 - 8 = Not relevant
    """
    rom, _unused1, _unused2 = \
        decode_command_register(self.COMMAND_REGISTER, 'ROM_PORT')
    self.ROM_PORT[rom] = self.ACCUMULATOR
    self.increment_pc(1)
    return self.ACCUMULATOR


def adm(self) -> Tuple[int, int]:
    """
    Name:           Add from memory with carry.

    Function:       The DATA RAM data character specified by the
                    last SRC instruction, plus the carry bit, are
                    added to the accumulator.
                    The data character is unaffected.
    Syntax:         ADM
    Assembled:      1110 1000
    Symbolic:       (M) + (ACC) + (CY) --> ACC, CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry/link is set to 1 if a sum greater than
                    MAX_4_BITS was generated to indicate a carry out;
                    otherwise, the carry/link is set to 0.
    """
    # Get value
    crb = self.read_current_ram_bank()
    chip, register, address = \
        decode_command_register(self.COMMAND_REGISTER, 'DATA_RAM_CHAR')
    absolute_address = convert_to_absolute_address(
        self, crb, chip, register, address)
    # Perform addition
    self.ACCUMULATOR = (self.ACCUMULATOR + self.RAM[absolute_address] +
                        self.read_carry())
    # Check for carry bit set/reset when an overflow is detected
    # i.e. the result is more than a 4-bit number (MAX_4_BITS)
    check_overflow(self)
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def sbm(self) -> Tuple[int, int]:
    """
    Name:           Subtract DATA RAM from memory with borrow.

    Function:       The value of the DATA RAM character specified
                    by the last SRC instruction is subtracted from
                    the accumulator with borrow.
                    The data character is unaffected.
    Syntax:         SDM
    Assembled:      1110 1000
    Symbolic:       (M) + (ACC) + (CY) --> ACC, CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   This instruction sets the carry bit if the result
                    generates no borrow, and resets the carry bit if
                    the result generates a borrow.
    Notes:  1       A borrow from the previous subtraction is indicated
                    by the carry bit being equal to one at the beginning
                    of this instruction.

            2       No borrow from the previous subtraction is indicated
                    by the carry bit being equal to zero at the beginning
                    of this instruction.

            3       The subtract with borrow operation is actually
                    performed by complementing each bit of the data
                    character and adding the resulting value plus
                    the complement of the carry bit to the accumulator.

            4       When this instruction is used to subtract numbers
                    greater than 4 bits in lengthJ the carry bit must
                    be complemented by the program between each required
                    subtraction operation.
    """
    # Get value
    crb = self.read_current_ram_bank()
    chip, register, address = \
        decode_command_register(self.COMMAND_REGISTER, 'DATA_RAM_CHAR')
    absolute_address = convert_to_absolute_address(
        self, crb, chip, register, address)
    value = self.RAM[absolute_address]

    # Perform addition
    value_complement = int(ones_complement(value, 4), 2)
    carry_complement = self.read_complement_carry()
    self.ACCUMULATOR = (self.ACCUMULATOR + value_complement +
                        carry_complement)
    # Check for carry bit set/reset when an overflow is detected
    # i.e. the result is more than a 4-bit number (MAX_4_BITS)
    check_overflow(self)
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def wpm(self) -> Tuple[int, int]:
    """
    Name:           Write program RAM.

    Function:       This is a special instruction which may be used to write
                    the contents of the accumulator into a half byte of
                    program RAM, or read the contents of a half byte of program
                    RAM into a ROM input port where it can be accessed by a
                    program.
                    The Carry bit is unaffected.
    Syntax:         WPM
    Assembled:      1110 0011
    Symbolic:       (A) --> (PRAM)
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.

    Notes:  1       Two WPM instructions must always appear in close
                    succession; that is, each time one WPM instruction
                    references a half byte of program RAM as indicated by an
                    SRC address, another WPM must access the other half byte
                    before the SRC address is altered. An internal counter
                    keeps track of which half-byte is being accessed. If only
                    one WPM occurs, this counter will be out of sync with the
                    program and errors will occur. In this situation, a RESET
                    pulse must be used to re-initialize the machine.

            2       A WPM instruction requires an SRC address to access program
                    RAM. Whenever a WPM is executed, the DATA RAM which happens
                    to correspond to this SRC address will also be written.
                    If data needed later in the program is being held in such
                    a DATA RAM, the programmer must save it elsewhere before
                    executing the WPM instruction.


    """
    chip, register, addr = decode_command_register(
        decimal_to_binary(8, self.COMMAND_REGISTER),
        'DATA_RAM_CHAR')
    rambank = self.read_current_ram_bank()
    address = convert_to_absolute_address(self, rambank, chip, register, addr)
    # Get the value of the WPM Counter
    wpm_counter = read_wpm_counter(self)

    # Writing
    if self.ROM_PORT[14] == 1:
        # Write enabled, so store
        value = self.ACCUMULATOR
        if wpm_counter == 'LEFT':
            print(wpm_counter)
            value = self.ACCUMULATOR << 4
            self.PRAM[address] = value
            self.RAM[address] = value
        if wpm_counter == 'RIGHT':
            value = self.ACCUMULATOR
            self.RAM[address] = self.RAM[address] + value
            self.PRAM[address] = self.PRAM[address] + value

    # Reading
    if self.ROM_PORT[14] != 1:
        # read
        if wpm_counter == 'LEFT':
            self.ROM_PORT[14] = self.PRAM[address] >> 4 << 4
        if wpm_counter == 'RIGHT':
            value = self.ACCUMULATOR
            self.ROM_PORT[15] = self.PRAM[address] << 4 >> 4

    flip_wpm_counter(self)

    self.increment_pc(1)

    return self.ROM_PORT[14], self.ROM_PORT[15]





#####################################################
#  _ _  _    ___   ___  _  _                        #
# (_) || |  / _ \ / _ \| || |                       #
#  _| || |_| | | | | | | || |_   _ __   ___  _ __   #
# | |__   _| | | | | | |__   _| | '_ \ / _ \| '_ \  #
# | |  | | | |_| | |_| |  | |   | | | | (_) | |_) | #
# |_|  |_|  \___/ \___/   |_|   |_| |_|\___/| .__/  #
#                                           | |     #
#                                           |_|     #
#                                                   #
#####################################################

"""
Commands in this module.

            NOP -   NO OPERATION

 Abbreviations used in the descriptions of each instruction's actions:

            (    )      the content of
            -->         is transferred to
            ACC	        Accumulator (4-bit)
            CY	        Carry/link Flip-Flop
            ACBR	    Accumulator Buffer Register (4-bit)
            RRRR	    Index register address
            RPn	        Index register pair address
            PL	        Low order program counter Field (4-bit)
            PM	        Middle order program counter Field (4-bit)
            PH	        High order program counter Field (4-bit)
            ai	        Order i content of the accumulator
            CMi	        Order i content of the command register
            M	        RAM main character location
            MSi	        RAM status character i
            DB (T)	    Data bus content at time T
            Stack	    The 3 registers in the address register
                        other than the program counter

    Additional Abbreviations:
            ~           Inverse (1's complement)
            .           logical OR

"""


def nop(self) -> int:
    """
    Name:           No Operation.

    Function:       No operation performed.
    Syntax:         NOP
    Assembled:      0000 0000
    Symbolic:       Not applicable
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable
    """
    self.increment_pc(1)
    return self.PROGRAM_COUNTER



"""Accumulator methods."""

# Import system modules
import os
import sys
sys.path.insert(1, '..' + os.sep + 'src')

# Import typing library
from typing import Tuple  # noqa

from hardware.exceptions import ValueTooLargeForAccumulator  # noqa


def check_overflow(self) -> Tuple[int, int]:
    """
    Check if an overflow is detected.

    If the result is more than a 4-bit number(MAX_4_BITS),
    then an overflow is detected.

    If there is an overflow detected, set the carry bit,
    otherwise reset the carry bit.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    self.ACCUMULATOR
        The value of the accumulator after adjusting for overflow
    self.CARRY
        The carry bit

    Raises
    ------
    N/A

    Notes
    -----
    N/A

    """
    if self.ACCUMULATOR > self.MAX_4_BITS:
        self.ACCUMULATOR = self.ACCUMULATOR - self.MAX_4_BITS + 1
        self.set_carry()
    else:
        self.reset_carry()
    return self.ACCUMULATOR, self.CARRY


def read_acbr(self) -> int:
    """
    Return the value of the ACBR interim property.

    This is used for swapping out accumulator with
    register contents during an XCH instruction

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    ACBR
        The value of the ACBR interim property

    """
    return self.ACBR


def read_accumulator(self) -> int:
    """
    Return the value of the accumulator.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    ACCUMULATOR
        The 4-bit value of the accumulator

    """
    return self.ACCUMULATOR


def set_accumulator(self, value: int) -> int:
    """
    Insert a value into the Accumulator.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    value: int, mandatory
        The value to insert

    Returns
    -------
    value
        The value of the accumulator

    Raises
    ------
    ValueTooLargeForAccumulator

    Notes
    -----
    N/A

    """
    if value > self.MAX_4_BITS:
        raise ValueTooLargeForAccumulator(' Value: ' + str(value))
    self.ACCUMULATOR = value
    return value



#############################################################
#       _ _ _   __   __  _ _      _         _               #
#      (_) | | /  \ /  \| | |    (_)_ _  __| |_____ __      #
#      | |_  _| () | () |_  _|   | | ' \/ _` / -_) \ /      #
#  _   |_| |_| \__/ \__/  |_|    |_|_||_\__,_\___/_\_\      #
# | |_ ___    __ _ __ __ _  _ _ __ _  _| |__ _| |_ ___ _ _  #
# |  _/ _ \  / _` / _/ _| || | '  \ || | / _` |  _/ _ \ '_| #
#  \__\___/  \__,_\__\__|\_,_|_|_|_\_,_|_\__,_|\__\___/_|   #
#                                                           #
#############################################################

"""
Commands in this module.

            ADD -   ADD REGISTER TO ACCUMULATOR WITH CARRY
            SUB -   SUBTRACT REGISTER FROM ACCUMULATOR WITH BORROW
            LD  -   LOAD ACCUMULATOR
            XCH -   EXCHANGE REGISTER AND ACCUMULATOR


 Abbreviations used in the descriptions of each instruction's actions:

            (    )      the content of
            -->         is transferred to
            ACC	        Accumulator (4-bit)
            CY	        Carry/link Flip-Flop
            ACBR	    Accumulator Buffer Register (4-bit)
            RRRR	    Index register address
            RPn	        Index register pair address
            PL	        Low order program counter Field (4-bit)
            PM	        Middle order program counter Field (4-bit)
            PH	        High order program counter Field (4-bit)
            ai	        Order i content of the accumulator
            CMi	        Order i content of the command register
            M	        RAM main character location
            MSi	        RAM status character i
            DB (T)	    Data bus content at time T
            Stack	    The 3 registers in the address register
                        other than the program counter

    Additional Abbreviations:
            ~           Inverse (1's complement)
            .           logical OR

"""

# Import typing library
from typing import Tuple


def add(self, register: int) -> Tuple[int, int]:
    """
    Name:           Add index register to accumulator with carry.

    Function:       The 4 bit content of the designated index register
                    is added to the content of the accumulator with carry.
                    The result is stored in the accumulator. (Note this
                    means the carry bit is also added)
    Syntax:         ADD <register>
    Assembled:      1000 <RRRR>
    Symbolic:       (RRRR) + (ACC) + (CY) --> ACC, CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry/link is set to 1 if a sum greater than
                    MAX_4_BITS was generated to indicate a carry out;
                    otherwise, the carry/link is set to 0. The 4 bit
                    content of the index register is unaffected.
    """
    from hardware.suboperations.accumulator import check_overflow  # noqa

    self.ACCUMULATOR = (self.ACCUMULATOR + self.REGISTERS[register] +
                        self.read_carry())
    # Check for carry bit set/reset when an overflow is detected
    # i.e. the result is more than a 4-bit number (MAX_4_BITS)
    check_overflow(self)
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def sub(self, register: int) -> Tuple[int, int]:
    """
    Name:           Subtract index register from accumulator with borrow.

    Function:       The 4 bit content of the designated index register is
                    complemented (ones complement) and added to content of
                    the accumulator with borrow and the result is stored
                    in the accumulator.
    Syntax:         SUB <register>
    Assembled:      1001 <RRRR>
    Symbolic:       (ACC) + ~(RRRR) + (CY) --> ACC, CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   If a borrow is generated, the carry bit is set to 0,
                    otherwise, it is set to 1.
                    The 4 bit content of the index register is unaffected.
    """
    from hardware.suboperations.utility import ones_complement, binary_to_decimal   # noqa

    carry = self.read_complement_carry()

    self.ACCUMULATOR = (self.ACCUMULATOR + binary_to_decimal(
        ones_complement(self.REGISTERS[register], 4))
        + carry)
    # Check for carry bit set/reset when borrow (overflow) is detected
    # i.e. the result is more than a 4-bit number (MAX_4_BITS)
    if self.ACCUMULATOR > self.MAX_4_BITS:
        self.ACCUMULATOR = self.ACCUMULATOR - (self.MAX_4_BITS + 1)
        self.set_carry()
    else:
        self.reset_carry()
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def ld(self, register: int) -> int:
    """
    Name:           Load index register to Accumulator.

    Function:       The 4 bit content of the designated index register
                    (RRRR) is loaded into accumulator.
                    The previous contents of the accumulator are lost.
    Syntax:         LD <value>
    Assembled:      1010 <RRRR>
    Symbolic:       (RRRR) --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry bit is not affected.
    """
    self.ACCUMULATOR = self.REGISTERS[register]
    self.increment_pc(1)
    return self.ACCUMULATOR


def xch(self, register: int) -> Tuple[int, int]:
    """
    Name:           Exchange index register and accumulator.

    Function:       The 4 bit content of designated index register is
                    loaded into the accumulator. The prior content of the
                    accumulator is loaded into the designed register.
    Syntax:         XCH <register>
    Assembled:      1011 <RRRR>
    Symbolic:       (ACC) --> ACBR, (RRRR) --> ACC, (ACBR) --> RRRR
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry bit is not affected.

    """
    self.ACBR = self.ACCUMULATOR
    self.ACCUMULATOR = self.REGISTERS[register]
    self.insert_register(register, self.ACBR)
    self.increment_pc(1)
    self.ACBR = 0
    return self.ACCUMULATOR, self.REGISTERS



"""Register methods."""
# Import system modules
import os
import sys
sys.path.insert(1, '..' + os.sep + 'src')

from hardware.exceptions import InvalidRegister, InvalidRegisterPair, \
    ValueTooLargeForRegister, ValueTooLargeForRegisterPair  # noqa


def increment_register(self, register: int) -> int:
    """
    Increment the value in a register by 1.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    register: int, mandatory
        register to increment

    Returns
    -------
    self.REGISTERS[register]
        value of the register post increment

    Raises
    ------
    InvalidRegister

    Notes
    -----
    N/A

    """
    if register < 0 or register > (self.NO_REGISTERS - 1):
        raise InvalidRegister('Register: ' + str(register))

    self.REGISTERS[register] = self.REGISTERS[register] + 1
    if self.REGISTERS[register] > self.MAX_4_BITS:
        self.REGISTERS[register] = 0
    return self.REGISTERS[register]


def insert_register(self, register: int, value: int) -> int:
    """
    Insert a value into a specific register.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    register: int, mandatory
        The number of the register to use

    value: int, mandatory
        The value to insert

    Returns
    -------
    value
        The value of the register

    Raises
    ------
    InvalidRegister
    ValueTooLargeForRegister

    Notes
    -----
    N/A

    """
    if (register < 0) or (register > 15):
        raise InvalidRegister('Register: ' + str(register))

    if value > 15:
        raise ValueTooLargeForRegister('Register: ' + str(register) + ',Value: ' + str(value))  # noqa
    self.REGISTERS[register] = value
    return value


def insert_registerpair(self, registerpair: int, value: int) -> int:
    """
    Insert a value into a specific register.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    registerpair: int, mandatory
        The number of the register to insert

    value: int, mandatory
        The value to insert

    Returns
    -------
    value
        The value of the register pair

    Raises
    ------
    InvalidRegisterPair
    ValueTooLargeForRegisterPair

    Notes
    -----
    N/A

    """
    if 0 < registerpair > 7:
        raise InvalidRegisterPair('Register Pair: ' +
                                  str(registerpair))
    if value > 256:
        raise ValueTooLargeForRegisterPair('Register Pair: ' +
                                           str(registerpair) +
                                           ',Value: ' +
                                           str(value))
    # Convert a register pair into a base register for insertion
    base_register = registerpair * 2
    self.insert_register(base_register, (value >> 4) & 15)   # Bit-shift right and remove low bits   # noqa
    self.insert_register(base_register + 1, value & 15)      # Remove low bits                       # noqa
    return value


def read_all_registers(self) -> list:
    """
    Return the values of the Registers.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    REGISTERS
        The values of all the Registers

    """
    return self.REGISTERS


def read_register(self, register: int) -> int:
    """
    Read a specific register.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    register: int, mandatory
        The number of the register to read

    Returns
    -------
    self.REGISTERS[register]
        The value of the requested register

    Raises
    ------
    InvalidRegister

    Notes
    -----
    N/A

    """
    if (register < 0) or (register > 15):
        raise InvalidRegister('Register:' + str(register))

    return self.REGISTERS[register]


def read_registerpair(self, registerpair: int) -> int:
    """
    Read a specific register pair.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    registerpair: int, mandatory
        The number of the register pair to read

    Returns
    -------
    value
        The value of the requested register pair

    Raises
    ------
    InvalidRegister

    Notes
    -----
    N/A

    """
    if (registerpair < 0 or registerpair > 7):
        raise InvalidRegisterPair('Register Pair: ' +
                                  str(registerpair))
    # Convert a register pair into a base register for insertion
    base_register = registerpair * 2

    hi = self.read_register(base_register)       # High 4-bits
    lo = self.read_register(base_register + 1)   # Low 4-bits
    return (hi << 4) + lo   # Bit-shift left high value and add low value



"""RAM methods."""
# Import system modules
import os
import sys
sys.path.insert(1, '..' + os.sep + 'src')

from hardware.suboperations.other import decode_command_register  # noqa


def rdx(self, character) -> int:
    """
    Read RAM status character X.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the Processor containing the registers, accumulator etc

    character:
        RAM STATUS CHARACTER to read

    Returns
    -------
    self.ACCUMULATOR
        The value read from the specified RAM STATUS CHARACTER

    """
    crb = self.read_current_ram_bank()
    chip, register, _none = \
        decode_command_register(self.COMMAND_REGISTER, 'DATA_RAM_STATUS_CHAR')
    self.ACCUMULATOR = self.STATUS_CHARACTERS[crb][chip][register][character]
    self.increment_pc(1)
    return self.ACCUMULATOR


def read_all_ram(self) -> list:
    """
    Return the values of all the locations of RAM.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    RAM
        The values of all the locations of RAM

    """
    return self.RAM


def read_all_ram_ports(self) -> list:
    """
    Return the values of all the RAM ports.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    RAM_PORT
        The values of all the RAM ports

    """
    return self.RAM_PORT


def read_all_pram(self) -> list:
    """
    Return the values of all the locations of PRAM.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    PRAM
        The values of all the locations of PRAM

    """
    return self.PRAM


def read_all_status_characters(self) -> list:
    """
    Return the values of all the RAM status characters.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    STATUS_CHARACTERS
        The values of all the RAM status characters

    """
    return self.STATUS_CHARACTERS


def read_current_ram_bank(self) -> int:
    """
    Return the current RAM bank i.e. the one selected by the SRC.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    CURRENT_RAM_BANK
        The current RAM bank value

    """
    return self.CURRENT_RAM_BANK


def write_ram_status(self, char: int) -> bool:
    """
    Write to a RAM status character.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    char: int, mandatory
        specified status character

    Returns
    -------
    True
        if the value is set successfully

    Raises
    ------
    N/A

    Notes
    -----
    No error checking is done in this function
    All parameters cannot be out of range, since the functions to
    place them in various registers etc all have range checking built in .

    Eventually - there will be error checking here

    """
    value = self.read_accumulator()
    crb = self.read_current_ram_bank()

    chip, register, _none = \
        decode_command_register(self.COMMAND_REGISTER,
                                'DATA_RAM_STATUS_CHAR')
    self.STATUS_CHARACTERS[crb][chip][register][char] = value
    return True



##########################################################################
#                         _ _ _   __   __  _ _                           #
#                        (_) | | /  \ /  \| | |                          #
#                        | |_  _| () | () |_  _|                         #
#                        |_| |_| \__/ \__/  |_|                          #
#                   _                    _ _      _                      #
#                  (_)_ __  _ __  ___ __| (_)__ _| |_ ___                #
#                  | | '  \| '  \/ -_) _` | / _` |  _/ -_)               #
#                  |_|_|_|_|_|_|_\___\__,_|_\__,_|\__\___|               #
#                                                                        #
##########################################################################

"""
Commands in this module.

            FIM -   FETCH IMMEDIATE
            LDM -   LOAD ACCUMULATOR IMMEDIATE


 Abbreviations used in the descriptions of each instruction's actions:

            (    )      the content of
            -->         is transferred to
            ACC	        Accumulator (4-bit)
            CY	        Carry/link Flip-Flop
            ACBR	    Accumulator Buffer Register (4-bit)
            RRRR	    Index register address
            RPn	        Index register pair address
            PL	        Low order program counter Field (4-bit)
            PM	        Middle order program counter Field (4-bit)
            PH	        High order program counter Field (4-bit)
            ai	        Order i content of the accumulator
            CMi	        Order i content of the command register
            M	        RAM main character location
            MSi	        RAM status character i
            DB (T)	    Data bus content at time T
            Stack	    The 3 registers in the address register
                        other than the program counter

    Additional Abbreviations:
            ~           Inverse (1's complement)
            .           logical OR

"""


def fim(self, registerpair: int, value: int) -> list:
    """
    Name:           Fetched immediate from ROM.

    Function:       The 2nd word represents 8-bits of data
                    which are loaded into the designated index register pair.
    Syntax:         FIM
    Assembled:      0010 RRR0
                    DDDD2  DDDD1
    Symbolic:       DDDD --> RRR0, DDDD1 --> RRR1
    Execution:      2 words, 16-bit code and an execution time of 21.6 usec.
    Side-effects:   Not Applicable
    """
    self.insert_registerpair(registerpair, value)
    self.increment_pc(2)
    return self.REGISTERS


def ldm(self, operand: int) -> int:
    """
    Name:           Load Accumulator Immediate.

    Function:       The 4 bits of immediate data are loaded into
                    the accumulator.
    Syntax:         LDM <value>
    Assembled:      1101 <DDDD>
    Symbolic:       DDDD --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry bit is not affected.
    """
    self.ACCUMULATOR = operand
    self.increment_pc(1)
    return self.ACCUMULATOR



"""Utility Operations."""
# Import typing library
from typing import Tuple

from hardware.exceptions import AddressOutOf8BitRange, \
    IncompatibleChunkBit, \
    InvalidBitValue, InvalidChunkValue, \
    NotABinaryNumber, ValueOutOfRangeForBits  # noqa


def binary_to_decimal(binary: str) -> int:
    """
    Converts a string value(which must be in binary form) to a decimal value.

    Parameters
    ----------
    binary: str, mandatory
        a string which represents the binary value

    Returns
    -------
    The decimal value of the supplied binary value

    Raises
    ------
    NotABinaryNumber: When a non-binary number is supplied

    Notes
    -----
    N/A

    """
    if len(binary) == 0:
        binary = '<empty>'
        raise NotABinaryNumber('"' + binary + '"')

    if len(binary.replace('0', '').replace('1', '')) != 0:
        raise NotABinaryNumber('"' + binary + '"')

    # Convert binary to decimal
    return int(binary, 2)


def convert_decimal_to_n_bit_slices(bits: int, chunk: int,
                                    decimal: int, result: str = 'b') -> str:
    """
    Converts a decimal to several binary or decimal values of specific lengths.

    Parameters
    ----------
    bits: int, mandatory
        number of bits of the source data

    chunk: int, mandatory
        number of bits required per chunk

    decimal: int: mandatory
        decimal value to convert

    result: str: mandatory
        'd' will generate a decimal output
        'b' will generate a binary output

    Returns
    -------
    The binary value of the supplied decimal value.

    Raises
    ------
    IncompatibleChunkBit: When the chunks do not fit exactly within the bits
    InvalidBitValue: When a bit value of not 4, 8 or 12 is specified
    InvalidChunkValue: When a chunk value of not 4, 8 or 12 is specified
    ValueOutOfRangeForBits: If the value supplied is either negative or is
                             out of range of the number of bits requested

    Notes
    -----
    N/A

    """
    if (bits not in [2, 4, 8, 12]):
        raise InvalidBitValue(' Bits: ' + str(bits))

    if (chunk not in [2, 4, 8, 12]):
        raise InvalidChunkValue(' Chunk: ' + str(chunk))

    if bits % chunk != 0:
        raise IncompatibleChunkBit(' Bits: ' + str(bits) +
                                   ' Chunk: ' + str(chunk))

    if (decimal > ((2 ** bits) - 1)) or (decimal < 0):
        raise ValueOutOfRangeForBits(' Value: ' + str(decimal) +
                                     ' Bits: ' + str(bits))

    binary = decimal_to_binary(bits, decimal)
    chunks = [binary[i:i+chunk] for i in range(0, len(binary), chunk)]
    if result != 'b':
        decimals = []
        for element in chunks:
            decimals.append(binary_to_decimal(element))
        chunks = decimals
    return chunks


def convert_to_absolute_address(self, rambank: int, chip: int,
                                register: int, address: int) -> int:
    """
    Convert a rambank, chip, register and address to an absolute RAM address.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    rambank: integer, mandatory
        The currently selected RAM bank

    chip : integer, mandatory
        1 of 4 chips

    register : integer, mandatory
        1 of 4 registers

    address : integer, mandatory
        address within a page

    Returns
    -------
    absolute_address
        The address from 0 - 4095

    Raises
    ------
    N/A

    Notes
    -----
    N/A

    """
    absolute_address = (rambank * self.RAM_BANK_SIZE) + \
        (chip * self.RAM_CHIP_SIZE) + \
        (register * self.RAM_REGISTER_SIZE) + address
    return absolute_address


def decimal_to_binary(bits: int, decimal: int) -> str:
    """
    Converts a decimal value into a binary value of a specified bit length.

    Parameters
    ----------
    bits: int, mandatory
        number of bits required for the conversion

    decimal: int: mandatory
        decimal value to convert

    Returns
    -------
    The binary value of the supplied decimal value.

    Raises
    ------
    InvalidBitValue: When a bit value of not 4, 8 or 12 is specified
    ValueOutOfRangeForBits: If the value supplied is either negative or is
                             out of range of the number of bits requested

    Notes
    -----
    N/A

    """
    if (bits not in [2, 4, 8, 12]):
        raise InvalidBitValue(' Bits: ' + str(bits))

    if (decimal > ((2 ** bits) - 1)) or (decimal < 0):
        raise ValueOutOfRangeForBits(' Value: ' + str(decimal) +
                                     ' Bits: ' + str(bits))

    # Convert decimal to binary
    binary = bin(decimal)[2:].zfill(bits)
    return binary


def ones_complement(value: str, bits: int) -> str:
    """
    Converts a decimal into a one's compliment value of a specified bit length.

    Parameters
    ----------
    value: str: mandatory
        decimal value to convert

    bits: int, mandatory
        number of bits required for the conversion

    Returns
    -------
    The one's compliment binary value of the supplied decimal value.

    Raises
    ------
    InvalidBitValue: When a bit value of not 4, 8 or 12 is specified
    ValueOutOfRangeForBits: If the value supplied is either negative or is
                             out of range of the number of bits requested

    Notes
    -----
    N/A

    """
    if (bits not in [2, 4, 8, 12]):
        raise InvalidBitValue(' Bits: ' + str(bits))

    if (int(value) > ((2 ** bits) - 1)) or (int(value) < 0):
        raise ValueOutOfRangeForBits(' Value: ' + str(value) +
                                     ' Bits: ' + str(bits))

    # Perform a one's complement
    # i.e. invert all the bits

    binary = decimal_to_binary(bits, int(value))
    # binary = str(bin(value))[2:].zfill(bits)
    ones = ''
    for x in range(bits):
        if binary[x] == '1':
            ones = ones + '0'
        else:
            ones = ones + '1'
    return ones


def split_address8(address: int) -> Tuple[str, str]:
    """
    Split a supplied decimal address into 2 4-bit words.

    Parameters
    ----------
    address: int, mandatory
        An 8-bit address in decimal format

    Returns
    -------
    address_left: str
        left-most 4 bits

    address_right: str
        right-most 4 bits

    Raises
    ------
    N/A

    Notes
    -----
    N/A

    """
    if (address < 0) or (address > 255):
        raise AddressOutOf8BitRange('Address: ' + str(address))

    address_left = bin(address)[2:].zfill(8)[:4]
    address_right = bin(address)[2:].zfill(8)[4:]
    return address_left, address_right



"""Program Counter methods."""

# Import system modules
import os
import sys
sys.path.insert(1, '..' + os.sep + 'src')

from hardware.exceptions import ProgramCounterOutOfBounds  # noqa


def increment_pc(self, words: int) -> int:
    """
    Increment the Program Counter by a specific number of words.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    words: int, mandatory
        The number of words to increment the Program Counter by

    Returns
    -------
    self.PROGRAM_COUNTER
        The new value of the Program Counter

    Raises
    ------
    ProgramCounterOutOfBounds

    Notes
    -----
    N/A

    """
    if self.PROGRAM_COUNTER + words > self.MEMORY_SIZE_RAM:
        raise ProgramCounterOutOfBounds('Program counter attempted to be' +
                                        ' set to ' +
                                        str(self.PROGRAM_COUNTER + words))
    self.PROGRAM_COUNTER = self.PROGRAM_COUNTER + words
    return self.PROGRAM_COUNTER


def inc_pc_by_page(self, pc: int) -> int:
    """
    Retrieve the pc's new value after being incremented by a page.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    pc: int, mandatory
        Current value of Program Counter

    Returns
    -------
    pc
        The new value of the Program Counter

    Raises
    ------
    ProgramCounterOutOfBounds

    Notes
    -----
    This function DOES NOT MODIFY the program counter, simply
    calculates the new  value of the counter. It is up to the
    calling  function to determine what to  do with the value.

    """
    if pc + self.PAGE_SIZE > self.MEMORY_SIZE_RAM:
        raise ProgramCounterOutOfBounds('Program counter attempted to be' +
                                        ' set to ' + str(pc + self.PAGE_SIZE))
    # Point the program counter to 1 page on
    pc = pc + self.PAGE_SIZE - 1
    return pc


def is_end_of_page(self, address: int, word: int) -> bool:
    """
    Determine if an instruction is located at the end of a memory page.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    address: int, mandatory
        Base address of the instruction

    word: int, mandatory
        Number of words in the instruction

    Returns
    -------
    True if the instruction is at the end of the memory page
    False if the instruction is not at the end of the memory page

    Raises
    ------
    N/A

    Notes
    -----
    N/A

    """
    page = address // self.PAGE_SIZE
    location = address - (page * self.PAGE_SIZE)
    word = word - 1
    return (location - word) == self.PAGE_SIZE - 1


def read_program_counter(self) -> int:
    """
    Return the value of the program counter.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    PROGRAM_COUNTER
        The value of the program counter (current instruction)

    """
    return self.PROGRAM_COUNTER



##########################################################################
#                         _ _ _   __   __  _ _                            #
#                        (_) | | /  \ /  \| | |                           #
#                        | |_  _| () | () |_  _|                          #
#                        |_| |_| \__/ \__/  |_|                           #
#                          _                 _   _                        #
#                  ____  _| |__ _ _ ___ _  _| |_(_)_ _  ___               #
#                 (_-< || | '_ \ '_/ _ \ || |  _| | ' \/ -_)              #
#                 /__/\_,_|_.__/_| \___/\_,_|\__|_|_||_\___|              #
#                                                                         #
###########################################################################

"""
Commands in this module.

            BBL -   BRANCH BACK AND LOAD
            JMS -   JUMP TO SUBROUTINE


 Abbreviations used in the descriptions of each instruction's actions:

            (    )      the content of
            -->         is transferred to
            ACC	        Accumulator (4-bit)
            CY	        Carry/link Flip-Flop
            ACBR	    Accumulator Buffer Register (4-bit)
            RRRR	    Index register address
            RPn	        Index register pair address
            PL	        Low order program counter Field (4-bit)
            PM	        Middle order program counter Field (4-bit)
            PH	        High order program counter Field (4-bit)
            ai	        Order i content of the accumulator
            CMi	        Order i content of the command register
            M	        RAM main character location
            MSi	        RAM status character i
            DB (T)	    Data bus content at time T
            Stack	    The 3 registers in the address register
                        other than the program counter

    Additional Abbreviations:
            ~           Inverse (1's complement)
            .           logical OR

"""

# Import typing library
from typing import Tuple

from hardware.exceptions import ValueOutOfRangeForStack


def bbl(self, accumulator: int) -> Tuple[int, int]:
    """
    Name:           Branch back and load data to the accumulator.

    Function:       The program counter (address stack) is pushed down one
                    level.
                    Program control transfers to the next instruction
                    following the last jump to subroutine (JMS)
                    instruction.
                    The 4 bits of data DDDD stored in the OPA portion of
                    the instruction are loaded to the accumulator.
                    BBL is used to return from subroutine to main program.
    Syntax:         BBL
    Assembled:      0110 DDDD
    Symbolic:       (Stack) --> PL, PM, PH
                    DDDD --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable
    """
    address = self.read_from_stack()
    self.PROGRAM_COUNTER = address
    self.ACCUMULATOR = accumulator
    return self.PROGRAM_COUNTER, self.ACCUMULATOR


def jms(self, address: int) -> int:
    """
    Name:           Jump to Subroutine.

    Function:       The address of the next instruction in sequence following
                    JMS (return address) is saved in the push down stack.
                    Program control is transferred to the instruction located
                    at the 12 bit address (AAAA3,AAAA2,AAAA1). Execution of a
                    return instruction (BBL) will cause the saved address to
                    be pulled out of the stack, therefore, program control
                    is transferred to the next sequential instruction after
                    the last JMS.
                    The push down stack has 4 registers. One of them is used
                    as the program counter, therefore nesting of JMS can occur
                    up to 3 levels.
    Syntax:         JMS
    Assembled:      0101 AAAA3
                    AAAA2 AAAA1
    Symbolic:       AAAA1 --> PL,
                    AAAA2 --> PM,
                    AAAA3 --> PH
    Execution:      2 words, 16-bit code and an execution time of 21.6 usec.
    Side-effects:   Not Applicable

    """
    # Add number of words so return address is correct
    self.write_to_stack(self.PROGRAM_COUNTER + 2)
    if 0 <= address <= 4095:
        self.PROGRAM_COUNTER = address - 1
        return self.PROGRAM_COUNTER
    raise ValueOutOfRangeForStack(' Value: ' + str(address))



# Read Processor Methods


def read_all_registers(self):
    """
    Return the values of the Registers

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    REGISTERS
        The values of all the Registers
    """
    return self.REGISTERS


def read_all_command_registers(self):
    """
    Return the values of all of the Command Registers

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    COMMAND_REGISTERS
        The values of all the Command Registers
    """
    return self.COMMAND_REGISTERS


def read_all_ram(self):
    """
    Return the values of all the locations of RAM

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    RAM
        The values of all the locations of RAM
    """
    return self.RAM


def read_all_rom(self):
    """
    Return the values of all the locations of ROM

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    ROM
        The values of all the locations of ROM
    """
    return self.ROM


def read_all_pram(self):
    """
    Return the values of all the locations of PRAM

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    PRAM
        The values of all the locations of PRAM
    """
    return self.PRAM


def read_all_stack(self):
    """
    Return the complete  stack

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    STACK
        The stack as a list
    """
    return self.STACK


def read_accumulator(self):
    """
    Return the value of the accumulator

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    ACCUMULATOR
        The 4-bit value of the accumulator
    """
    return self.ACCUMULATOR


def read_all_rom_ports(self):
    """
    Return the values of all the ROM ports

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    ROM_PORT
        The values of all the ROM ports
    """
    return self.ROM_PORT


def read_all_ram_ports(self):
    """
    Return the values of all the RAM ports

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    RAM_PORT
        The values of all the RAM ports
    """
    return self.RAM_PORT


def read_current_ram_bank(self):
    """
    Return the current RAM bank i.e. the one selected  by the SRC

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    CURRENT_RAM_BANK
        The current RAM bank value
    """
    return self.CURRENT_RAM_BANK


def read_carry(self):
    """
    Return the value of the carry flag

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    CARRY
        A 1 or a 0 (the value of the carry flag)
    """
    return self.CARRY


def read_pin10(self):
    """
    Return the value of PIN 10 on the i4004 chip (simulated test pin)

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    PIN_10_SIGNAL_TEST
        The value of the simulated test pin
    """
    return self.PIN_10_SIGNAL_TEST


def read_wpm_counter(self):
    """
    Return the value of the WPM counter ("LEFT" or "RIGHT") - internally
    keeps track of which 4-bits of a register will be transferred with the
    WPM instruction

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    WPM_COUNTER
        The value of the WPM counter
    """
    return self.WPM_COUNTER


def read_acbr(self):
    """
    Return the value of the ACBR interim property used for
    swapping out accumulator with register contents during
    an XCH instruction

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    ACBR
        The value of the ACBR interim property
    """
    return self.ACBR


def read_program_counter(self):
    """
    Return the value of the program counter

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    PROGRAM_COUNTER
        The value of the program counter (current instruction)
    """
    return self.PROGRAM_COUNTER


def read_stack_pointer(self):
    """
    Return the value of the stack pointer

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    STACK_POINTER
        The value of the stack pointer used to determine where in the stack
        return addresses to  subroutines (for example) are
    """
    return self.STACK_POINTER


def read_all_status_characters(self):
    """
    Return the values of all the RAM status characters

    Parameters
    ----------
    self : processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    STATUS_CHARACTERS
        The values of all the RAM status characters
    """
    return self.STATUS_CHARACTERS



"""Carry methods."""


def read_carry(self) -> int:
    """
    Return the value of the carry flag.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    CARRY
        A 1 or a 0 (the value of the carry flag)

    """
    return self.CARRY


def read_complement_carry(self) -> int:
    """
    Reads the complement of the carry bit, but doesn't change the value.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
        The complement of the carry bit

    Raises
    ------
    N/A

    Notes
    -----
    N/A

    """
    return 1 if self.CARRY == 0 else 0


def reset_carry(self) -> int:
    """
    Resets the carry bit.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    self.CARRY
        The carry bit

    Raises
    ------
    N/A

    Notes
    -----
    N/A

    """
    # Reset the carry bit
    self.CARRY = 0
    return self.CARRY


def set_carry(self) -> int:
    """
    Set the carry bit.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    self.CARRY
        The carry bit

    Raises
    ------
    N/A

    Notes
    -----
    N/A

    """
    # Set the carry bit
    self.CARRY = 1
    return self.CARRY



##########################################################################
#                         _ _ _   __   __  _ _                           #
#                        (_) | | /  \ /  \| | |                          #
#                        | |_  _| () | () |_  _|                         #
#                        |_| |_| \__/ \__/  |_|                          #
#                                                _        _              #
#         _ __  ___ _ __  ___ _ _ _  _   ___ ___| |___ __| |_            #
#        | '  \/ -_) '  \/ _ \ '_| || | (_-</ -_) / -_) _|  _|           #
#        |_|_|_\___|_|_|_\___/_|  \_, | /__/\___|_\___\__|\__|           #
#                                 |__/                                   #
##########################################################################

"""
Commands in this module.

            DCL -   DESIGNATE COMMAND LINE
            SRC -   SEND REGISTER CONTROL


 Abbreviations used in the descriptions of each instruction's actions:

            (    )      the content of
            -->         is transferred to
            ACC	        Accumulator (4-bit)
            CY	        Carry/link Flip-Flop
            ACBR	    Accumulator Buffer Register (4-bit)
            RRRR	    Index register address
            RPn	        Index register pair address
            PL	        Low order program counter Field (4-bit)
            PM	        Middle order program counter Field (4-bit)
            PH	        High order program counter Field (4-bit)
            ai	        Order i content of the accumulator
            CMi	        Order i content of the command register
            M	        RAM main character location
            MSi	        RAM status character i
            DB (T)	    Data bus content at time T
            Stack	    The 3 registers in the address register
                        other than the program counter

    Additional Abbreviations:
            ~           Inverse (1's complement)
            .           logical OR


"""

from hardware.exceptions import InvalidRamBank, InvalidRegisterPair
from hardware.suboperations.utility import decimal_to_binary


def dcl(self) -> int:
    """
    Name:           Designate command line.

    Function:       The content of the three least significant accumulator
                    bits is transferred to the command control register
                    within the CPU. This instruction provides RAM bank
                    selection when multiple RAM banks are used.
                    (If no DCL instruction is sent out, RAM Bank number
                    zero is automatically selected after application of at
                    least one RESET). See below for RAM Bank selection table.
                    DCL remains latched until it is changed.
    Syntax:         DCL
    Assembled:      1111 1101
    Symbolic:       a0 --> CM0, a1 --> CM1, a2 --> CM2
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable

                    The selection is made according to the following table.
                    (ACC)	CM-RAMi                     Enabled	Bank No.
                    -----	------------------------	------------
                    X000	CM-RAM0	                    Bank 0
                    X001	CM-RAM1	                    Bank 1
                    X010	CM-RAM2	                    Bank 2
                    X100	CM-RAM3	                    Bank 3
                    X011	CM-RAM1, CM-RAM1	        Bank 4
                    X101	CM-RAM1, CM-RAM3	        Bank 5
                    X110	CM-RAM2, CM-RAM3	        Bank 6
                    X111	CM-RAM1, CM-RAM2, CM-RAM3	Bank 7
    """
    if self.ACCUMULATOR > 7:
        raise InvalidRamBank('RAM bank : ' + str(self.ACCUMULATOR))

    self.CURRENT_RAM_BANK = self.ACCUMULATOR
    self.increment_pc(1)
    return self.CURRENT_RAM_BANK


def src(self, registerpair: int) -> int:
    """
    Name:           Send register control.

    Function:       The 8 bit content of the designated index register pair
                    is sent to the RAM address register at X2 and X3.
                    A subsequent read, write, or I/O operation of the RAM will
                    utilize this address. Specifically, the first 2 bits of the
                    address designate a RAM chip; the second 2 bits designate
                    1 out of 4 registers within the chip;
                    the last 4 bits designate 1 out of 16 4-bit main memory
                    characters within the register.

                    This command is also used to designate a ROM for a
                    subsequent ROM I/O port operation. The first 4 bits
                    designate the ROM chip number to be selected. The address
                    in ROM or RAM is not cleared until the next SRC
                    instruction is executed.
                    The 8 bit content of the index register is unaffected.
    Syntax:         SRC
    Assembled:      0010 RRR1
    Symbolic:       (RRRO) --> DB (X2)
                    (RRR1) --> DB (X3)
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec..
    Side-effects:   Not Applicable
    """
    from hardware.suboperation import decimal_to_binary
    if registerpair > 7:
        raise InvalidRegisterPair('Register pair : ' + str(registerpair))

    self.increment_pc(1)
    address = self.read_registerpair(registerpair)
    self.COMMAND_REGISTER = decimal_to_binary(8, address)
    return address



"""Stack methods."""
# Import system modules
import os
import sys
sys.path.insert(1, '..' + os.sep + 'src')

from hardware.exceptions import ValueOutOfRangeForStack  # noqa


def read_all_stack(self) -> list:
    """
    Return the complete stack.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    STACK
        The stack as a list

    """
    return self.STACK


def read_from_stack(self) -> int:
    """
    Read from the stack.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    value
        the 12-bit number read from the stack

    Raises
    ------
    N/A

    Notes
    -----
    The stack pointer begins at 2, and then moves toward 0

        First Read             Second Read          Third Read
        +------------+       +------------+        +------------+
        |     d      | <--SP |      d     |        |      d     |
        |     b      |       |      b     |        |      b     | <---SP
        |     c      |       |      c     | <--SP  |      c     |
        +------------+       +------------+        +------------+

    """
    if self.STACK_POINTER == 2:
        self.STACK_POINTER = 0
    else:
        self.STACK_POINTER = self.STACK_POINTER + 1
    value = self.STACK[self.STACK_POINTER]

    return value


def read_stack_pointer(self) -> int:
    """
    Return the value of the stack pointer.

    Parameters
    ----------
    self : Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    Returns
    -------
    STACK_POINTER
        The value of the stack pointer used to determine where in the stack
        return addresses to  subroutines (for example) are

    """
    return self.STACK_POINTER


def write_to_stack(self, value: int) -> bool:
    """
    Write to the stack.

    Parameters
    ----------
    self: Processor, mandatory
        The instance of the processor containing the registers, accumulator etc

    value: int, mandatory
        value to write to the stack

    Returns
    -------
    True
        if the value is written to the stack successfully

    Raises
    ------
    ValueOutOfRangeForStack

    Notes
    -----
    The stack pointer begins at 2, and then moves toward 0

      After 2 writes         After 3 writes         After 3 writes
      +------------+         +------------+        +------------+
      |     a      |         |      a     | <--SP  |      d     |
      |     b      |         |      b     |        |      b     | <---SP
      |            | <--SP   |      c     |        |      c     |
      +------------+         +------------+        +------------+

    After 3 writes, address "a" is lost

    """
    if (value < 0 or value > 4095):
        raise ValueOutOfRangeForStack(' Value: ' + str(value))

    self.STACK[self.STACK_POINTER] = value
    self.STACK_POINTER = self.STACK_POINTER - 1
    if self.STACK_POINTER == -1:
        self.STACK_POINTER = 2
    return True



"""Definition of an i4004 processor."""


class Processor:

    """Functionality and chracteristics of an i4004 processor."""

    #  pylint: disable=import-outside-toplevel
    # Turn off import-outside-toplevel warning for this class
    from hardware import opcodes

    # Import opcode mechanics
    from hardware.instructions.accumulator import clb, clc, cma, cmc, \
        daa, dac, iac, kbp, ral, rar, stc, tcc, tcs
    from hardware.instructions.idx import fin, inc
    from hardware.instructions.idxacc import add, ld, sub, xch
    from hardware.instructions.immediate import fim, ldm
    from hardware.instructions.io_ram import adm, rdm, rdr, rd0, rd1, rd2, \
        rd3, sbm, wmp, wpm, wrm, wrr, wr0, wr1, wr2, wr3
    from hardware.instructions.memory_select import dcl, src
    from hardware.instructions.nop import nop
    from hardware.instructions.subroutine import bbl, jms
    from hardware.instructions.transfer_control import isz, jcn, jin, jun

    from hardware.suboperations.other import decode_command_register, \
        read_all_command_registers
    from hardware.suboperations.utility import binary_to_decimal, \
        convert_decimal_to_n_bit_slices, convert_to_absolute_address, \
        decimal_to_binary, ones_complement, split_address8

    # Import suboperations
    from hardware.suboperations.accumulator import check_overflow, \
        read_acbr, read_accumulator, set_accumulator
    from hardware.suboperations.carry import read_carry, \
        read_complement_carry, reset_carry, set_carry
    from hardware.suboperations.init import init_command_registers, \
        init_pram, init_ram, init_registers, init_rom, init_stack, \
        init_wpm_counter
    from hardware.suboperations.pc import inc_pc_by_page, increment_pc, \
        is_end_of_page, read_program_counter
    from hardware.suboperations.pin10 import read_pin10, write_pin10
    from hardware.suboperations.ram import rdx, read_all_pram, read_all_ram, \
        read_all_ram_ports, read_all_status_characters, \
        read_current_ram_bank, write_ram_status
    from hardware.suboperations.registers import increment_register, \
        insert_register, insert_registerpair, read_all_registers, \
        read_register, read_registerpair
    from hardware.suboperations.rom import read_all_rom, read_all_rom_ports
    from hardware.suboperations.stack import read_all_stack, read_from_stack, \
        read_stack_pointer, write_to_stack
    from hardware.suboperations.wpm import flip_wpm_counter, read_wpm_counter
    #  pylint: enable=import-outside-toplevel

    # Operations to read the processor components
    # Some used internally,

    # i4004 Processor characteristics
    MAX_4_BITS = 15             # Maximum value 4 bits can hold

    MEMORY_SIZE_RAM = 4096      # Number of 4-bit words in RAM
    MEMORY_SIZE_ROM = 4096      # Number of 4-bit words in ROM
    MEMORY_SIZE_PRAM = 4096     # Number of 4-bit words in PRAM
    PAGE_SIZE = 256             # Number of 4-bit words in a memory page
    STACK_SIZE = 3              # Number of 12-bit registers in the stack
    NO_REGISTERS = 16           # Number of registers
    NO_ROM_PORTS = 16           # Number of ROM output ports
    NO_CHIPS_PER_BANK = 4       # Number of memory chips per Data RAM Bank
    RAM_BANK_SIZE = 256         # Size in 4-bit addresses of a Data RAM Bank
    RAM_CHIP_SIZE = 64          # Size in 4-bit addresses of a single RAM chip
    RAM_REGISTER_SIZE = 16      # Number of 4-bit registers in a RAM chip
    MSB = 8                     # Most significant bit of a 4 bit register
    NO_DRB = 8                  # Number of Data RAM Banks (0-7)
    NO_COMMAND_REGISTERS = 8    # Number of command registers
    NO_STATUS_REGISTERS = 4     # Number of Status registers per memory chip
    NO_STATUS_CHARACTERS = 4    # Number of Status chars per status register

    # Instruction table
    INSTRUCTIONS = opcodes.instructions.opcodes

    # Initialise processor

    def __init__(self):
        """Initialise an instance of the processor."""
        # Set up all the internals of the processor
        self.COMMAND_REGISTERS = []  # Command Register (Select Data RAM Bank)

        # Set up RAM
        self.RAM = []                                # RAM
        self.RAM_PORT = [[0 for _bank in range(4)]   # RAM Ports
                         for _chip in range(8)]
        # Set up ROM
        self.ROM = []                                # ROM
        self.ROM_PORT = [0 for _bank in range(self.NO_ROM_PORTS)]   # ROM ports

        self.PRAM = []               # Program RAM
        self.REGISTERS = []          # Registers (4-bit)
        self.STACK = []              # The stack - 3 x 12-bit registers

        self.COMMAND_REGISTER = 0

        # Set up RAM status characters
        self.STATUS_CHARACTERS = [[[[0 for _char in range(4)]
                                    for _reg in range(4)]
                                   for _chip in range(4)]
                                  for _bank in range(self.NO_DRB)]

        # Creation of processor simulated hardware
        # Pin 10 on the physical chip is the "test" pin
        # and can be read by the JCN instruction
        self.PIN_10_SIGNAL_TEST = 0
        self.write_pin10(0)

        # Initialise Internals
        self.set_accumulator(0)      # Initialise the accumulator
        self.ACBR = 0                # Accumulator Buffer Register
        self.STACK_POINTER = 2       # Stack Pointer
        self.PROGRAM_COUNTER = 0     # Program Counter - 12-bit value

        self.init_stack()
        self.init_command_registers()
        self.init_registers()
        self.init_pram()
        self.init_ram()
        self.init_rom()
        self.CURRENT_DRAM_BANK = 0   # Current Data RAM Bank
        self.CURRENT_RAM_BANK = 0    # Current Program RAM Bank
        self.reset_carry()           # Reset the carry bit
        self.init_wpm_counter()      # WPM Counter (Left/Right flip)

    #  END OF PROCESSOR DEFINITION



##########################################################################
#                                                                        #
#  _ _ _   __   __  _ _                              _      _            #
# (_) | | /  \ /  \| | |   __ _ __ __ _  _ _ __ _  _| |__ _| |_ ___ _ _  #
# | |_  _| () | () |_  _| / _` / _/ _| || | '  \ || | / _` |  _/ _ \ '_| #
# |_| |_| \__/ \__/  |_|  \__,_\__\__|\_,_|_|_|_\_,_|_\__,_|\__\___/_|   #
#                                                                        #
##########################################################################

"""
Commands in this module.

            CLB -   CLEAR BOTH
            CLC -   CLEAR CARRY
            CMA -   COMPLEMENT ACCUMULATOR
            CMC -   COMPLEMENT CARRY
            DAA -   DECIMAL ADJUST ACCUMULATOR
            DAC -   DECREMENT ACCUMULATOR
            IAC -   INCREMENT ACCUMULATOR
            KBP -   KEYBOARD PROCESS
            RAL -   ROTATE ACCUMULATOR LEFT THROUGH CARRY
            RAR -   ROTATE ACCUMULATOR RIGHT THROUGH CARRY
            STC -   SET CARRY
            TCC -   TRANSMIT CARRY AND CLEAR
            TCS -   TRANSFER CARRY SUBTRACT


 Abbreviations used in the descriptions of each instruction's actions:

        (    )      the content of
        -->         is transferred to
        ACC	        Accumulator (4-bit)
        CY	        Carry/link Flip-Flop
        ACBR	    Accumulator Buffer Register (4-bit)
        RRRR	    Index register address
        RPn	        Index register pair address
        PL	        Low order program counter Field (4-bit)
        PM	        Middle order program counter Field (4-bit)
        PH	        High order program counter Field (4-bit)
        ai	        Order i content of the accumulator
        CMi	        Order i content of the command register
        M	        RAM main character location
        MSi	        RAM status character i
        DB (T)	    Data bus content at time T
        Stack	    The 3 registers in the address register
                     other than the program counter

 Additional Abbreviations:
        ~           Inverse (1's complement)
        .           logical OR

"""

# Import typing library
from typing import Tuple, Any


def clb(self: Any) -> Tuple[int, int]:
    """
    Name:           Clear Both.

    Function:       Set accumulator and carry/link to 0.
    Syntax:         CLB
    Assembled:      1111 0000
    Symbolic:       0 --> ACC, 0 --> CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable
    """
    self.set_accumulator(0)
    self.reset_carry()
    self.increment_pc(1)
    return self.read_accumulator(), self.read_carry()


def clc(self: Any) -> int:
    """
    Name:           Clear Carry.

    Function:       Set carry/link to 0.
    Syntax:         CLC
    Assembled:      1111 0001
    Symbolic:       0 --> CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable
    """
    self.reset_carry()
    self.increment_pc(1)
    return self.read_carry()


def cma(self: Any) -> int:
    """
    Name:           Complement Accumulator.

    Function:       The content of the accumulator is complemented.
                    The carry/link is unaffected.
    Syntax:         CMA
    Assembled:      1111 0100
    Symbolic:       ~a3 ~a2 ~a1 ~a0 --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable
    """
    from hardware.suboperations.utility import ones_complement  # noqa

    ones_acc = int(ones_complement(self.ACCUMULATOR, 4), 2)
    self.ACCUMULATOR = ones_acc
    self.increment_pc(1)
    return self.ACCUMULATOR


def cmc(self: Any) -> int:
    """
    Name:           Complement carry.

    Function:       The carry/link content is complemented.
    Syntax:         CLC
    Assembled:      1111 0011
    Symbolic:       ~(CY) --> CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable
    """
    if self.CARRY == 1:
        self.reset_carry()
    else:
        self.set_carry()
    self.increment_pc(1)
    return self.CARRY


def daa(self: Any) -> Tuple[int, int]:
    """
    Name:           Decimal adjust accumulator.

    Function:       The accumulator is incremented by 6 if
                    either the carry/link is 1 or if the accumulator
                    content is greater than 9.
    Syntax:         DAA
    Assembled:      1111 1011
    Symbolic:       (ACC) + (0000 or 0110) --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry/link is set to a 1 if the result generates
                    a carry, otherwise it is unaffected.
    """
    if (self.read_carry() == 1 or self.ACCUMULATOR > 9):
        self.ACCUMULATOR = self.ACCUMULATOR + 6
        if self.ACCUMULATOR > self.MAX_4_BITS:
            self.ACCUMULATOR = self.ACCUMULATOR - self.MAX_4_BITS - 1
            self.set_carry()
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def dac(self: Any) -> Tuple[int, int]:
    """
    Name:           Decrement accumulator.

    Function:       The content of the accumulator is decremented by 1.
    Syntax:         DAC
    Assembled:      1111 1000
    Symbolic:       (ACC) -1 --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   A borrow sets the carry/link to 0;
                    No borrow sets the carry/link to a 1.
    """
    self.ACCUMULATOR = self.ACCUMULATOR + 15
    if self.ACCUMULATOR >= self.MAX_4_BITS:
        self.ACCUMULATOR = self.MSB
        self.set_carry()
    else:
        self.reset_carry()
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def iac(self: Any) -> Tuple[int, int]:
    """
    Name:           Increment accumulator.

    Function:       The content of the accumulator is incremented by 1.
    Syntax:         IAC
    Assembled:      1111 0010
    Symbolic:       (ACC) +1 --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   No overflow sets the carry/link to 0;
                    overflow sets the carry/link to a 1.
    """
    self.ACCUMULATOR = self.ACCUMULATOR + 1
    if self.ACCUMULATOR == self.MAX_4_BITS + 1:
        self.ACCUMULATOR = 0
        self.set_carry()
    else:
        self.reset_carry()
    self.increment_pc(1)
    return self.read_accumulator(), self.read_carry()


def kbp(self: Any) -> int:
    """
    Name:           Keyboard process.

    Function:       A code conversion is performed on the accumulator content,
                    from 1 out of n to binary code.
                    If the accumulator content has more than one bit on, the
                    accumulator will be set to 15.
                    (to indicate error) - conversion table below
    Syntax:         KBP
    Assembled:      1111 1100
    Symbolic:       (ACC) --> KBP, ROM --> ACC
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry/link is unaffected.


                       (ACC)           (ACC)
                       before          after
                        KBP	  	        KBP
                        0000	---->	0000
                        0001	---->	0001
                        0010	---->	0010
                        0100	---->	0011
                        1000	---->	0100
                        0011	---->	1111    Error
                        0101	---->	1111    Error
                        0110	---->	1111    Error
                        0111	---->	1111    Error
                        1001	---->	1111    Error
                        1010	---->	1111    Error
                        1011	---->	1111    Error
                        1100	---->	1111    Error
                        1101	---->	1111    Error
                        1110	---->	1111    Error
                        1111	---->	1111    Error
    """
    acc = self.read_accumulator()
    self.increment_pc(1)
    if acc in (0, 1, 2):
        return acc

    if acc == 4:
        self.set_accumulator(3)
        return self.read_accumulator()

    if acc == 8:
        self.set_accumulator(4)
        return self.read_accumulator()

    # Error
    self.set_accumulator(15)
    return self.read_accumulator()


def ral(self: Any) -> Tuple[int, int]:
    """
    Name:           Rotate left.

    Function:       The content of the accumulator and carry/link
                    are rotated left.
    Syntax:         RAL
    Assembled:      1111 0101
    Symbolic:       C0 --> a0, a(i) --> a(i+1), a3 -->CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry bit will be set to the highest significant
                    bit of the accumulator.
    """
    c0 = self.read_carry()
    # Shift left
    self.ACCUMULATOR = self.ACCUMULATOR * 2
    # Set carry bit correctly
    if self.ACCUMULATOR >= self.MAX_4_BITS:
        self.set_carry()
    else:
        self.reset_carry()
    # If necessary remove non-existent bit 5
    if self.ACCUMULATOR > self.MAX_4_BITS:
        self.ACCUMULATOR = self.ACCUMULATOR - self.MAX_4_BITS - 1
    # Add original carry bit
    self.ACCUMULATOR = self.ACCUMULATOR + c0
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def rar(self: Any) -> Tuple[int, int]:
    """
    Name:           Rotate right.

    Function:       The content of the accumulator and carry/link
                    are rotated right.
    Syntax:         RAR
    Assembled:      1111 0110
    Symbolic:       a0 --> CY, a(i) --> a(i-1), C0 -->a3
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry bit will be set to the lowest significant
                    bit of the accumulator.
    """
    c0 = self.read_carry()
    # Set carry bit correctly
    if self.ACCUMULATOR % 2 == 0:
        self.reset_carry()
    else:
        self.set_carry()
    # Shift right
    self.ACCUMULATOR = self.ACCUMULATOR // 2
    # Add carry to high-order bit of accumulator
    self.ACCUMULATOR = self.ACCUMULATOR + (c0 * self.MSB)
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def stc(self: Any) -> int:
    """
    Name:           Set Carry.

    Function:       Set carry/link to 1.
    Syntax:         STC
    Assembled:      1111 1010
    Symbolic:       1 --> CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   Not Applicable
    """
    self.set_carry()
    self.increment_pc(1)
    return self.read_carry()


def tcc(self: Any) -> Tuple[int, int]:
    """
    Name:           Transmit carry and clear.

    Function:       The accumulator is cleared.
                    The least significant position of the accumulator
                    is set to the value of the carry/link.
    Syntax:         TCC
    Assembled:      1111 0111
    Symbolic:       0 --> ACC, (CY) --> a0, 0 --> CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry bit will be reset..
    """
    self.set_accumulator(0)
    self.ACCUMULATOR = self.read_carry()
    self.reset_carry()
    self.increment_pc(1)
    return self.ACCUMULATOR, self.CARRY


def tcs(self: Any) -> Tuple[int, int]:
    """
    Name:           Transfer Carry Subtract.

    Function:       The accumulator is set to 9 if the carry/link is 0.
                    The accumulator is set to 10 if the carry/link is a 1.
    Syntax:         TCS
    Assembled:      1111 1001
    Symbolic:       1001 --> ACC if (CY) = 0
                    1010 --> ACC if (CY) = 1
                    0 --> CY
    Execution:      1 word, 8-bit code and an execution time of 10.8 usec.
    Side-effects:   The carry/link is set to 0.
    """
    if self.read_carry() == 0:
        self.set_accumulator(9)
    else:
        self.set_accumulator(10)
    self.reset_carry()
    self.increment_pc(1)
    return self.read_accumulator(), self.read_carry()
